#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2017, Ignacio Fdez. Galván                             *
#***********************************************************************

# Test the accuracy of the fitting polynomials generated by the genfit.py script.
# The roots and weights are computed at the "data" points and in the midpoints.

from mpmath import mp
from rys_aux import *
import sys

# Set floating point precision
# Single precision (32 bits): 24-bit significand, 6 decimal places
# Double precision (64 bits): 53-bit significand, 15 decimal places
# Quadruple precision (128 bits): 113-bit significand, 33 decimal places
mp.dps = 33

# Read the data from stdin, expected to be a file formated
# as a Molcas "rysrw" file
# Maximum number of roots and degree of fitting polynomials
nmax, fit_degree = (int(i) for i in sys.stdin.readline().split())
# Requested accuracy (not needed)
accuracy = []
while (len(accuracy) < nmax):
  accuracy.extend([mp.mpf(i) for i in sys.stdin.readline().split()])
# Value of x from which the asymptotic approximation is accurate enough (not needed)
xmax = []
while (len(xmax) < nmax):
  xmax.extend([mp.mpf(i) for i in sys.stdin.readline().split()])
# Step for the x values
dx = []
while (len(dx) < nmax):
  dx.extend([mp.mpf(i) for i in sys.stdin.readline().split()])
# Number of x values
nx = []
while (len(nx) < nmax):
  nx.extend([int(i) for i in sys.stdin.readline().split()])
# Number of fitting intervals
npol = []
while (len(npol) < nmax):
  npol.extend([int(i) for i in sys.stdin.readline().split()])
# Map for which fitting interval to use for each x value
# (values are 1-based, so subtract 1)
polmap = [[] for i in range(nmax)]
# And center of each interval (offset value for x)
center = [[] for i in range(nmax)]
for n in range(nmax):
  while (len(polmap[n]) < nx[n]):
    polmap[n].extend([int(i)-1 for i in sys.stdin.readline().split()])
  while (len(center[n]) < npol[n]):
    center[n].extend([mp.mpf(i) for i in sys.stdin.readline().split()])
# Coefficients for all roots and weights and intervals
coeff = [[] for i in range(nmax)]
for n in range(nmax):
  # First read a single list
  tmp = []
  while (len(tmp) < npol[n]*(n+1)*(fit_degree+1)*2):
    tmp.extend([mp.mpf(i) for i in sys.stdin.readline().split()])
  # Then arrange it, the nesting order in the file is:
  #   roots, weights
  #     0th...6th order coefficients
  #       root 1...n
  #         interval 1...npol
  m = 0
  for j in range(2):
    coeff[n].append([])
    for k in range(fit_degree+1):
      coeff[n][j].append([])
      for l in range(n+1):
        coeff[n][j][k].append([])
        coeff[n][j][k][l] = tmp[m:m+npol[n]]
        m += npol[n]

# Now it's time to actually test the fittings for each order of Rys polynomials
for n in range(nmax):
  max_error = [zero, zero]
  # Test all the "data points" (x values) and mid-points too
  for ix in range(2*nx[n]-1):
    x = ix*dx[n]/2

    # Compute roots and weights
    roots, weights = rysroots(n+1, x)

    # Select approximating polynomial
    ipol = polmap[n][int(mp.floor(x/dx[n]))]
    xoff = x-center[n][ipol]
    # For each root and weight calculate the value from the fitting polynomial
    roots_fit = []
    weights_fit = []
    for i in range(n+1):
      pol = []
      for j in range(fit_degree+1):
        pol.insert(0, coeff[n][0][j][i][ipol])
      roots_fit.append(mp.polyval(pol, xoff))
      pol = []
      for j in range(fit_degree+1):
        pol.insert(0, coeff[n][1][j][i][ipol])
      weights_fit.append(mp.polyval(pol, xoff))
    # Obtain the maximum relative error
    error_r = max([abs((a-b)/b) for a,b in zip(roots_fit, roots)])
    error_w = max([abs((a-b)/b) for a,b in zip(weights_fit, weights)])
    error = max(error_r, error_w)
    # Keep separate track of maximum error at data points and mid-points
    i = ix%2
    max_error[i] = max(max_error[i], error)
  print(n+1, figs(max_error[0], exp=True), figs(max_error[1], exp=True))



# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2017, Ignacio Fdez. Galván                             *
#***********************************************************************
#===============================================================================
# Auxiliary functions for the rysroots tools
#===============================================================================

from mpmath import mp
from decimal import Decimal

# Useful constants
zero  = mp.mpf(0)
one   = mp.mpf(1)
two   = mp.mpf(2)
four  = mp.mpf(4)
quart = one/four
pi    = mp.pi

#===============================================================================
# Wrapper function to return an mp float converted to string, with the specified
# number of decimal places. "exp" can be a boolean to force exponential notation
# or an integer to specify the minimum number of digits in the exponent.
def figs(x, n=20, exp=0, sign=False):
  if (sign and x >= 0.0):
    s = ' '
  else:
    s = ''
  # mp always returns '0.0' for zero. Bypass it using standard formats
  if (x == 0.0):
    fmt = '{{0:.{0}f}}'.format(n)
    # python uses two-digit exponent, but mp uses the least amount. Hard code it for zero
    if (exp):
      fmt += 'e+'+'0'*exp
    return s+fmt.format(0.0)
  # Convert significant figures to decimal places:
  # With exponent, integer part should be 1 figure, so it's n+1
  # Without exponent, floor(log10(abs(x)))+1 gives the number of integer figures (except for zero)
  if (exp):
    tmp = s+mp.nstr(mp.mpf(x), n+1, strip_zeros=False, min_fixed=mp.inf, max_fixed=-mp.inf, show_zero_exponent=True)
    # Add zeros to the exponent if necessary
    exppos = tmp.find('e')+2
    diff = exp-(len(tmp)-exppos)
    if (diff > 0):
      tmp = tmp[:exppos] + '0'*diff + tmp[exppos:]
    return tmp
  else:
    return s+mp.nstr(mp.mpf(x), int(mp.floor(mp.log10(abs(x))))+n+1, strip_zeros=False, min_fixed=-mp.inf, max_fixed=mp.inf)
#===============================================================================

#===============================================================================
# Function to print a list in chunks of n elements
# The argument f is a function that converts each element to string
def printlist(l, n, f=str):
  for i in range(0, len(l), n):
    num = min(n, len(l)-i)
    print(' '.join(map(f, l[i:i+num])))
#===============================================================================

#===============================================================================
# range analogue for real numbers (with exact short decimal representation)
def frange(start, end, step):
  # Use a parallel Decimal representation to keep operations exact,
  # but yield the original representation.
  # This could mean that the final value is slightly larger than "end"
  a = Decimal(str(start))
  b = Decimal(str(end))
  c = Decimal(str(step))
  if (step < 0.0):
    f = -1
  else:
    f = 1
  x,d = (start,a)
  i = 0
  while (f*d <= f*b):
    yield x
    i += 1
    x,d = (start+i*step, a+i*c)
#===============================================================================

#===============================================================================
# Given alpha and beta recursion coefficients for a set of orthogonal
# polynomials, compute the corresponding roots and weights for a Gaussian
# quadrature. The roots and weights are obtained from the eigenvalues and
# eigenvectors of a tridiagonal matrix containing the alpha and beta
# coefficients (Jacobi matrix).
#
# Code translated directly from:
#   Walter Gautschi. ACM Trans. Math. Softw. 20 (1994) 21-62
#   doi:10.1145/174603.174605
# (routine supplied by G. H. Golub according to a footnote)
def gauss(alpha, beta, eps, nloop=30):
  n = len(alpha)
  if (len(beta) != n):
    raise

  # initialize
  root = alpha[:]
  weight = [one]+[zero for i in range(n-1)]
  aux = list(map(mp.sqrt, beta[1:]))
  aux.append(zero)

  for l in range(n):
    for j in range(nloop):
      for m in range(l, n):
        if (m == n-1):
          break
        if (abs(aux[m]) <= eps*(abs(root[m])+abs(root[m+1]))):
          break
      dp = root[l]
      if (m == l):
        break
      dg = (root[l+1]-dp)/(two*aux[l])
      dr = mp.sqrt(dg*dg+one)
      if (mp.sign(dg) < 0):
        dr *= -1
      dg = root[m]-dp+aux[l]/(dg+dr)
      ds = one
      dc = one
      dp = zero
      mml = m-l
      for ii in range(1, mml+1):
        i = m-ii
        df = ds*aux[i]
        db = dc*aux[i]
        if (abs(df) < abs(dg)):
          ds = df/dg
          dr = mp.sqrt(ds*ds+one)
          aux[i+1] = dg*dr
          dc = one/dr
          ds = ds*dc
        else:
          dc = dg/df
          dr = mp.sqrt(dc*dc+one)
          aux[i+1] = df*dr
          ds = one/dr
          dc = dc*ds
        dg = root[i+1]-dp
        dr = (root[i]-dg)*ds+two*dc*db
        dp = ds*dr
        root[i+1] = dg+dp
        dg = dc*dr-db
        df = weight[i+1]
        weight[i+1] = ds*weight[i]+dc*df
        weight[i] = dc*weight[i]-ds*df
      root[l] = root[l]-dp
      aux[l] = dg
      aux[m] = zero
    if (j >= nloop-1):
      raise

  # scale, sort and return
  return (list(t) for t in zip(*sorted(zip(root, [beta[0]*w*w for w in weight]))))
#===============================================================================

#===============================================================================
# Generate alpha and beta recursion coefficients according to the discretized
# Stieltjes procedure for the orthogonal polynomials with an inner product
# given by the input roots and weights
#
# Code translated directly from:
#   Walter Gautschi. ACM Trans. Math. Softw. 20 (1994) 21-62
#   doi:10.1145/174603.174605
def stieltjes(n, roots, weights):
  naux = len(roots)
  if (len(weights) != naux):
    raise
  if (n > naux):
    raise

  huge = mp.mpf('1.0e+60')
  tiny = mp.mpf('1.0e-60')
  beta = [sum(weights)]
  alpha = [sum([x*w for x,w in zip(roots, weights)])/beta[0]]

  p0 = [zero for i in range(naux)]
  p1 = [zero for i in range(naux)]
  p2 = [one for i in range(naux)]
  sum0 = beta[0]
  for k in range(1, n):
    sum1 = zero
    sum2 = zero
    for m in range(naux):
      if (weights[m] == zero):
        continue
      p0[m] = p1[m]
      p1[m] = p2[m]
      p2[m] = (roots[m]-alpha[k-1])*p1[m]-beta[k-1]*p0[m]
      if ((abs(p2[m]) > huge) or (abs(sum2) > huge)):
        raise
      t = weights[m]*p2[m]*p2[m]
      sum1 = sum1+t
      sum2 = sum2+t*roots[m]
    if (abs(sum1) < tiny):
      raise
    alpha.append(sum2/sum1)
    beta.append(sum1/sum0)
    sum0 = sum1

  return alpha, beta
#===============================================================================

#===============================================================================
# Generate alpha and beta recursion coefficients according to the Lanczos
# orthogonal reduction method. This is an alternative (more robust, slower)
# to Stieltjes's procedure
#
# Code translated directly from:
#   Walter Gautschi. ACM Trans. Math. Softw. 20 (1994) 21-62
#   doi:10.1145/174603.174605
def lanczos(n, roots, weights):
  naux = len(roots)
  if (len(weights) != naux):
    raise
  if (n > naux):
    raise

  p0 = roots[:]
  p1 = [zero for i in range(naux)]
  p1[0] = weights[0]
  for i in range(1, naux):
    gam = one
    sig = zero
    t = zero
    p_i = weights[i]
    xlam = roots[i]
    for k in range(i+1):
      rho = p1[k]+p_i
      tmp = gam*rho
      tsig = sig
      if (rho <= zero):
        gam = one
        sig = zero
      else:
        gam = p1[k]/rho
        sig = p_i/rho
      tk = sig*(p0[k]-xlam)-gam*t
      p0[k] = p0[k]-(tk-t)
      t = tk
      if (sig <= zero):
        p_i = tsig*p1[k]
      else:
        p_i = t*t/sig
      tsig = sig
      p1[k] = tmp

  return p0[0:n], p1[0:n]
#===============================================================================

#===============================================================================
# Compute alpha and beta three-term recursion parameters for Rys polynomials.
# An auxiliary shifted Legendre quadrature is used, for which roots and weights
# are computed the first time this is run. Note that naux=300 is most probably
# overkill.
def rysab(n, x, naux=300, preci=mp.mpf('1e-20'), method='s'):
  # Generate roots and weights for the auxiliary quadrature (shifted Legendre)
  #   alpha(i) = 1/2
  #   beta(0)  = 1
  #   beta(i)  = 1 / (4 - 1/i^2)
  # Do this only the first time (results saved as attributes) or if
  # the "naux" or "preci" parameters change
  if ((rysab.roots_aux == None) or
      (rysab.naux != naux) or (rysab.preci != preci)):
    rysab.naux = naux
    rysab.preci = preci
    alpha_aux = [mp.mpf('0.5') for i in range(naux)]
    beta_aux = [one]
    for i in range(1, naux):
      beta_aux.append(quart/(four-one/mp.mpf(i*i)))
    rysab.roots_aux, rysab.weights_aux = gauss(alpha_aux, beta_aux, preci)

  rtsgau = [r*r for r in rysab.roots_aux]
  wtsgau = [w*mp.exp(-x*r) for r,w in zip(rtsgau, rysab.weights_aux)]
  if (method == 's'):
    return stieltjes(n, rtsgau, wtsgau)
  elif (method == 'l'):
    return lanczos(n, rtsgau, wtsgau)
  else:
    raise
rysab.roots_aux = None
#===============================================================================

#===============================================================================
# Compute roots and weights for Rys polynomials
def rysroots(n, x, preci=mp.mpf('1e-20'), **kwargs):
  kwargs.update({'preci': preci})
  alpha, beta = rysab(n, x, **kwargs)
  return tuple(gauss(alpha, beta, preci))
#===============================================================================



# -*- coding: utf-8 -*-
#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015,2018,2020, Ignacio Fdez. Galván                   *
#***********************************************************************

try:
  from docutils.parsers.rst import Directive
  do_directive = True
except ImportError:
  do_directive = False
import re
import codecs
import os.path
from help_text_format import *

X_clean = re.compile(r'^%+\+*')
X_line = re.compile(r'(<\/?(help|key|group|select|module|emil|command|alternate|include).*?>)', flags=re.IGNORECASE)
X_inhelp = re.compile(r'<help>', flags=re.IGNORECASE)
X_inhelp2 = re.compile(r'<\/help>', flags=re.IGNORECASE)
X_comment = re.compile(r'\s*<!--.*?-->')
X_tag = re.compile(r'<')
X_help = re.compile(r'\s*(<\/?help)', flags=re.IGNORECASE)
X_key = re.compile(r'(<\/?(key|command))', flags=re.IGNORECASE)
X_group = re.compile(r'(<\/?(group|select|include))', flags=re.IGNORECASE)
X_alt = re.compile(r'(<\/?alternate)', flags=re.IGNORECASE)
X_pretty = re.compile(r'(?<=.)(?=<[^/])')

H_head = re.compile(r'%%(keyword|description):', flags=re.IGNORECASE)
H_remove = re.compile(r'(%|<!--$|-->$)')
H_keyw = re.compile(r'%%keyword:\s*(.*?)\s*<(.*?)>\s*(.*)', flags=re.IGNORECASE)
H_desc = re.compile(r'%%description', flags=re.IGNORECASE)

nowrap = re.compile(r'$')
nowrap2 = re.compile(br'\u00a6'.decode('raw_unicode_escape'))

# Create the xmldoc directive
#
if (do_directive):
  class XMLDocDirective(Directive):
    has_content = True

    def run(self):
      self.assert_has_content()
      # Add a global "XMLDocs" attribute to the environment
      env = self.state.document.settings.env
      if not hasattr(env, 'XMLDocs'):
        env.XMLDocs = []
      # The "XMLDocs" attribute contains all the xmldoc pieces
      env.XMLDocs.append({
        'docname': env.docname,
        'lineno': self.lineno,
        'content': [line for line in self.content],
      })
      return []

# Write all the XML pieces at the end
#
def write_XMLDocs(app, exception):
  env = app.builder.env
  if hasattr(env, 'XMLDocs'):
    data_dir = app.config.data_dir
    with codecs.open(os.path.join(data_dir, 'keyword.xml'), 'w', 'utf-8') as keywordsfile:
      keywordsfile.write('<!-- This file is generated automatically from the OpenMolcas documentation -->\n')
      keywordsfile.write('<ROOT>\n')
      # Sort by docname, then by lineno
      docs = list(set([piece['docname'] for piece in env.XMLDocs]))
      for doc in sorted(docs):
        for piece in sorted([pc for pc in env.XMLDocs if pc['docname'] == doc], key=lambda x: x['lineno']):
          # Reformat the text for XML
          text = reformat_XML(piece['content'])
          if (text):
            keywordsfile.write('\n'.join(text)+'\n')
      keywordsfile.write('</ROOT>\n')

# Reformat before writing to the XML
#
def reformat_XML(piece):
  # Remove lines we do not want in the XML, and do some initial cleanup
  text = ' '.join([nowrap.sub(br'\u00a6'.decode('raw_unicode_escape'),X_clean.sub('',line)) for line in piece if not H_head.match(line)])
  # XML tags in separate lines
  text = X_line.sub(r'\n\1\n', text)
  text = text.split('\n')
  inhelp = False
  for i in range(len(text)):
    line = text[i].strip()
    if (X_inhelp.match(line)):
      help_text = ''
      inhelp = True
    if (X_inhelp2.match(line)):
      # Reformat
      if (inhelp and not text[i-1]):
        help_doc = parse_help_text(help_text)
        if (False):
          # Format as HTML-like
          text[i-1] = X_pretty.sub(r'\n', ET.tostring(help_doc, encoding='unicode'))
        else:
          # Format as plain text
          text[i-1] = format_help_text(help_doc)
      inhelp = False
    # Restore newlines inside <HELP>, discard text lines outside
    if (not X_tag.match(line)):
      if (inhelp):
        line = nowrap2.sub(r'\n', line).strip('\n')
        help_text += line
      line = ''
    # Indent XML tags
    line = X_help.sub(r'         \1', line)
    line = X_alt.sub(r'         \1', line)
    line = X_key.sub(r'      \1', line)
    line = X_group.sub(r'   \1', line)
    text[i] = line
  # Remove empty lines
  text = [nowrap2.sub(r' ',line) for line in filter(bool, text)]
  return text

# Write all the xmldoc pieces at the end
#
def write_Help(app, exception):
  env = app.builder.env
  if hasattr(env, 'XMLDocs'):
    data_dir = app.config.data_dir
    with codecs.open(os.path.join(data_dir, 'keyword.db'), 'w', 'utf-8') as keywordsfile:
      keywordsfile.write('#This file is generated automatically from the OpenMolcas documentation\n')
      keywordsfile.write('#\n')
      # Sort by docname, then by lineno
      docs = list(set([piece['docname'] for piece in env.XMLDocs]))
      for doc in sorted(docs):
        keywordsfile.write('[%s]\n' % os.path.basename(doc).upper())
        for piece in sorted([pc for pc in env.XMLDocs if pc['docname'] == doc], key=lambda x: x['lineno']):
          # Reformat the text for the Help file
          text = reformat_Help(piece['content'])
          if (text):
            keywordsfile.write('\n'.join(text)+'\n')

def reformat_Help(piece):
  # Find the start of the help section
  index = None
  header = None
  for i in range(len(piece)):
    if (H_head.match(piece[i])):
      index = i
      header = help_header(piece[index])
      break
  if (index is None):
    text = ''
  else:
    # Remove lines we do not want in the help file, and do some initial cleanup
    text = ' '.join([nowrap.sub(br'\u00a6'.decode('raw_unicode_escape'),X_clean.sub('',line)) for line in ['\n'] + piece[index:] if not (H_remove.match(line) or X_line.match(line))])
  text = text.split('\n')
  for i in range(len(text)):
    line = text[i].strip()
    # Restore newlines
    line = nowrap2.sub(r'\n', line).strip('\n')
    text[i] = line
  # Remove empty lines
  text = [nowrap2.sub(r' ',line) for line in filter(bool, text)]
  # Reformat
  help_doc = parse_help_text('\n'.join(text))
  text = [format_help_text(help_doc)]
  if text == ['']: text = []
  if (header): text.insert(0, header)
  return text

def help_header(line):
  if (H_desc.match(line)):
    return ':Description:D'
  elif (H_keyw.match(line)):
    tmp = H_keyw.match(line)
    typ = tmp.group(2)[0].upper()
    if (typ == 'C'): typ = 'B'
    gui = tmp.group(3)
    if (not gui): gui = ' '
    return ':%s:%s:%s:' % (tmp.group(1), typ, gui)
  else:
    return ''

# Purge the xmldoc pieces from a file when it is changed
#
def purge_XMLDocs(app, env, docname):
  if hasattr(env, 'XMLDocs'):
    env.XMLDocs = [piece for piece in env.XMLDocs
                   if piece['docname'] != docname]

# Setup
#
if (do_directive):
  def setup(app):
    app.add_directive('xmldoc', XMLDocDirective)
    app.connect('env-purge-doc', purge_XMLDocs)
    app.connect('build-finished', write_XMLDocs)
    app.connect('build-finished', write_Help)
    app.add_config_value('data_dir', 'xmldoc_files', '')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015,2018, Ignacio Fdez. Galván                        *
#***********************************************************************

'''
This packs pymolcas into a single-file executable python package that can be
copied around. It still uses python and it's trivial to recover the original
files, comments included. Obfuscation is not the goal here, just getting
something that's easy to run, move and distribute.
'''

from __future__ import (unicode_literals, division, absolute_import, print_function)
try:
  from builtins import bytes
except ImportError:
  from future.builtins import bytes

import glob, os.path

from setuptools import setup

files = [os.path.splitext(x)[0] for x in glob.glob('*.py')]
files.remove('export')
files.remove('pack')
files.remove('imports')

setup(
  script_args=['bdist_wheel', '-d', '.'],
  py_modules=files,
)

# Now try to make the package runnable:
# Add the self-running trick code to the top of the file,
# rename it and make it executable

import sys, os, stat

exe_name = 'pymolcas.whl'
magic_code = '''#!/bin/sh
name=`readlink -f "$0"`
exec {0} -c "import sys, os; sys.path.insert(0, '$name'); from pymolcas import main; sys.exit(main(my_name='$name'))" "$@"
'''.format(sys.executable)

wheel = glob.glob('*.whl')[0]
with open(exe_name, 'wb') as new:
  new.write(bytes(magic_code, 'ascii'))
  with open(wheel, 'rb') as original:
    data = True
    while (data):
      data = original.read(4096)
      new.write(data)
os.remove(wheel)
st = os.stat(exe_name)
os.chmod(exe_name, st.st_mode | stat.S_IEXEC)



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2021, Ignacio Fdez. Galván                             *
#***********************************************************************

import sys
import re
import argparse
import numpy as np
import h5py

version = '1.0'

################################################################################
# FUNCTIONS
################################################################################

# Copy dataset name_in of file_in into dataset name_out of file_out,
# but attributes (description) are taken from name_out.
# If size is specified, no data is copied and a dataset is created with the new size
def copy_dataset(file_in, file_out, name_in, name_out=None, size=None):
  if name_out is None:
    name_out = name_in
  no_attrs = name_in != name_out
  if size is None:
    file_in.copy(file_in[name_in], file_out, name_out, without_attrs=no_attrs)
  else:
    file_out.create_dataset(name_out, size, dtype=file_in[name_out].dtype)
  if no_attrs:
    for attr in file_in[name_out].attrs.keys():
      file_out[name_out].attrs.create(attr, file_in[name_out].attrs[attr])

# Rearrange a symmetry-blocked array into a square array
def sblock_to_square(data, nbas, triangular=False):
  if len(nbas) == 1:
    if triangular:
      mtrx = np.zeros((nbas[0], nbas[0]))
      ii, jj = np.tril_indices(nbas[0])
      mtrx[ii,jj] = data
      mtrx[jj,ii] = data
    else:
      mtrx = np.reshape(data, (nbas[0], nbas[0]))
  else:
    nb = int(np.sum(nbas))
    mtrx = np.zeros((nb, nb))
    j = 0
    for i,n in enumerate(nbas):
      ni = int(np.sum(nbas[:i]))
      if triangular:
        tmp = np.zeros((n, n))
        ii, jj = np.tril_indices(n)
        tmp[ii,jj] = data[j:j+n*(n+1)//2]
        tmp[jj,ii] = data[j:j+n*(n+1)//2]
        mtrx[np.ix_(range(ni,ni+n), range(ni,ni+n))] = tmp[:,:]
        j += n*(n+1)//2
      else:
        mtrx[np.ix_(range(ni,ni+n), range(ni,ni+n))] = np.reshape(data[j:j+n**2], (n, n))
        j += n**2
  return mtrx

# Normalize a rotation matrix,
# i.e., find the closest unitary matrix
def normalize_rot(A, test=True):
  U, s, Vh = np.linalg.svd(A)
  if test:
    if not np.allclose(s, 1.0, atol=1e-3):
      sys.exit('Not a unitary rotation')
  return U @ Vh

# Returns unique permutations (permutations with repeated elements).
# From https://stackoverflow.com/questions/6284396/permutations-with-unique-values
def permute_unique(nums):
  perms = [[]]
  for n in nums:
    new_perm = []
    for perm in perms:
      for i in range(len(perm) + 1):
        new_perm.append(perm[:i]+[n]+perm[i:])
        # handle duplication
        if i < len(perm) and perm[i] == n:
          break
    perms = new_perm
  return perms

# Returns a list of Cartesian factors (0=x, 1=y, 2=z), given l and m
# (note that m goes from -l to +l and beyond)
def factors(l, m):
  ly = int(np.floor((np.sqrt(8*(m+l)+1)-1)/2))
  lz = m+l-ly*(ly+1)//2
  lx = l-ly
  ly -= lz
  f = []
  for i in range(lx):
    f.append(0)
  for i in range(ly):
    f.append(1)
  for i in range(lz):
    f.append(2)
  return f

# Compute the rotation matrix for the functions in a Cartesian shell
# of angular momentum l, given the real-space rotation matrix R.
# (The cases l=0,1 are trivial, but we don't treat them separately)
def cartesian_shell_rotation(l, R):
  assert l >= 0
  num = (l+1)*(l+2)//2
  Rot = np.zeros((num, num))
  # Pre-compute the Cartesian factors for each m
  f = []
  for m in range(-l, num-l):
    f.append(factors(l, m))
  for m1 in range(num):
    perms = permute_unique(f[m1])
    for m2 in range(num):
      # Each element [m2,m1] is of the form sum(prod(R[i,j])),
      # where the sum extends over all distinct permutations of the m1 factors
      # and the prod is over the factors of m1 & m2
      for p in perms:
        K = 1.0
        for i,j in zip(p, f[m2]):
          K *= R[i,j]
        Rot[m2,m1] += K
  return Rot

# Compute the rotation matrix for the functions in a spherical harmonics shell
# of angular momentum l, given the real-space rotation matrix R (implicit).
# Uses the recursive algorithm of Ivanic and Ruedenberg (doi:10.1021/jp953350u, doi:10.1021/jp9833350),
# which is valid for l > 1.
# Note that this builds the inverse/transpose, so indexing may look reversed.
def spherical_shell_rotation(l):
  assert l > 1
  def rP(i,mu,m2):
    if abs(mu) > l-1:
      return 0
    # Instead of referring to the original rotation matrix R, we use RotM[1],
    # which is already reordered according to m (but indices are 0,1,2, not -1,0,1).
    # NB: The first two cases are wrong in the original (m' -> l-1)
    elif m2 == l:
      return RotM[1][2,i+1]*RotM[l-1][2*l-2,l-1+mu]-RotM[1][0,i+1]*RotM[l-1][0,l-1+mu]
    elif m2 == -l:
      return RotM[1][2,i+1]*RotM[l-1][0,l-1+mu]+RotM[1][0,i+1]*RotM[l-1][2*l-2,l-1+mu]
    else:
      return RotM[1][1,i+1]*RotM[l-1][l-1+m2,l-1+mu]
  num = 2*l+1
  Rot = np.zeros((num, num))
  for m2 in range(-l, l+1):
    if abs(m2) == l:
      den = 2*l*(2*l-1)
    else:
      den = (l+m2)*(l-m2)
    for m1 in range(-l, l+1):
      u = np.sqrt((l+m1)*(l-m1)/den)
      rU = rP(0,m1,m2)
      if m1 == 0:
        v = -0.5*np.sqrt(2*(l-1)*l/den)
        rV = rP(1,1,m2)+rP(-1,-1,m2)
        w = 0
        rW = 0
      else:
        s = 1 if m1 > 0 else -1
        absm = abs(m1)
        v = 0.5*np.sqrt((l+absm-1)*(l+absm)/den)
        # NB: The case m < 0 is wrong in the correction (sqrt(1-delta) -> sqrt(1+delta))
        if absm == 1:
          rV = np.sqrt(2)*rP(s,absm-1,m2)
        else:
          rV = rP(1,m1-s,m2)-s*rP(-1,-m1+s,m2)
        w = -0.5*np.sqrt((l-absm-1)*(l-absm)/den)
        rW = rP(1,m1+s,m2)+s*rP(-1,-m1-s,m2)
      Rot[l+m2,l+m1] = u*rU+v*rV+w*rW
  return Rot

################################################################################

# Command-line arguments
parser = argparse.ArgumentParser(formatter_class=lambda prog: argparse.HelpFormatter(prog,max_help_position=24,width=120))
parser.version = f'MORT: molecular orbital rotation and translation (version {version})'
parser.add_argument('-v', '--version', help='print version number', action='version')
parser.add_argument('-r', '--rotate', help='rotation matrix to apply (9 real values)', type=float, nargs=9, metavar=('Rxx','Rxy','Rxz','Ryx','Ryy','Ryz','Rzx','Rzy','Rzz'))
parser.add_argument('-t', '--translate', help='translation vector to apply after rotation (3 real values, in bohr)', type=float, nargs=3, metavar=('Tx','Ty','Tz'))
parser.add_argument('-x', '--exchange', help='reorder atoms (comma-separated list of integers giving order in the output file)', metavar='n1,n2,...')
parser.add_argument('-d', '--desymmetrize', help='desymmetrize data', action='store_true')
parser.add_argument('-a', '--all', help='copy all unhandled datasets and attributes to output file (some items may be wrong!)', action='store_true')
parser.add_argument('infile', help='input file (HDF5 format)', metavar='input_file')
parser.add_argument('outfile', help='output file (HDF5 format, will be overwritten)', metavar='output_file')
if len(sys.argv) < 2 and sys.stdin.isatty():
  parser.print_help(sys.stderr)
  sys.exit(1)
else:
  args = vars(parser.parse_args())

# Read input file

with h5py.File(args['infile'], 'r') as f:
  nsym = f.attrs['NSYM']
  if nsym != 1:
    if not args['desymmetrize']:
      sys.exit('Error: Symmetry is only supported with --desymmetrize')
  try:
    if nsym == 1:
      b_id_ref = f['BASIS_FUNCTION_IDS'][:]
    else:
      b_id_ref = f['DESYM_BASIS_FUNCTION_IDS'][:]
    nb = len(b_id_ref)
  except KeyError:
    sys.exit('Error: Input file does not have basis set information')
  try:
    if nsym == 1:
      nat = f.attrs['NATOMS_UNIQUE']
    else:
      nat = f.attrs['NATOMS_ALL']
  except KeyError:
    sys.exit('Error: Input file does not have number of atoms')

# Compute transformation matrices

if args['rotate']:
  do_rot = True
  R = np.reshape(args['rotate'], (3,3))
  # Normalize to clean up numerical errors
  R = normalize_rot(R)

  # Maximum l in spherical shells
  # (Cartesian shells have negative l and don't need recursion)
  max_l = np.max(b_id_ref[:,2])

  # Group the basis functions by shells
  # (same center, l, and radial part)
  l_idx = {}
  b_num = []
  for i,b in enumerate(b_id_ref):
    l = b[2] 
    for j,bb in enumerate(b_id_ref[:i,:]):
      if tuple(b[0:3]) == tuple(bb[0:3]):
        num = b_num[j] 
        l_idx[l][num].append(i)
        b_num.append(num)
        break
    else:
      if l in l_idx:
        l_idx[l].append([i])
      else:
        l_idx[l] = [[i]]
      b_num.append(len(l_idx[l])-1)
    
  # Make sure each shell is ordered by m
  for l in l_idx:
    for i,shell in enumerate(l_idx[l]):
      l_idx[l][i] = sorted(shell, key=lambda x: b_id_ref[x,3])

  # Compute the rotation matrices for the different shells
  RotM = {}
  # First the Cartesian shells
  for l in l_idx.keys():
    # s and p shells (l=0,1) are always considered Cartesian
    if l < 2:
      RotM[l] = cartesian_shell_rotation(abs(l), R)
  # Reorder the p rotation to match the m ordering (1=x, -1=y, 0=z)
  RotM[1] = RotM[1][np.ix_([1,2,0],[1,2,0])]
  # Then the spherical harmonics shells
  for l in range(2,max_l+1):
    RotM[l] = spherical_shell_rotation(l)
  
  # Build the big (all functions), sparse transformation matrix and its inverse,
  # which is needed because the Cartesian rotations are not unitary
  U = np.zeros((nb, nb))
  V = np.zeros((nb, nb))
  for l in l_idx:
    for shell in l_idx[l]:
      U[np.ix_(shell,shell)] = RotM[l]
      if l < 0:
        V[np.ix_(shell,shell)] = np.linalg.inv(RotM[l])
      else:
        V[np.ix_(shell,shell)] = RotM[l].T
else:
  do_rot = False
  RotM = {}
  R = np.eye(3)
  U = np.eye(nb)
  V = np.eye(nb)

if args['translate']:
  do_trans = True
  T = np.array(args['translate'])
else:
  do_trans = False
  T = np.zeros(3)

if args['exchange']:
  do_exch = True
  # List that specifies how the centers are reordered, its inverse, and the corresponding matrix
  reindex = [int(i)-1 for i in args['exchange'].split(',')]
  assert len(reindex) == nat, 'Error: length of --exchange list does not match number of atoms'
  reindex_inv = []
  for i in range(len(reindex)):
    reindex_inv.append(reindex.index(i))
  reorder = np.eye(nat)[reindex]
else:
  do_exch = False
  reindex = list(range(nat))
  reindex_inv = reindex
  reorder = np.eye(nat)

if args['desymmetrize']:
  do_desym = nsym != 1
else:
  do_desym = False

if do_desym or do_exch:
  # List that specifies how basis functions are reordered and its matrix
  bf_reindex = []
  # Create an auxiliary array to use lexsort on:
  aux_bfid = b_id_ref.copy()
  for b in aux_bfid:
    if do_exch:
      b[0] = reindex_inv[b[0]-1]+1
    b[2] = abs(b[2])
    # p shells are ordered x,y,z
    if b[2] == 1:
      if b[3] == 1:
        b[3] = -1
      elif b[3] == -1:
        b[3] = 0
      elif b[3] == 0:
        b[3] = 1
  bf_reindex = np.lexsort((aux_bfid[:,1], aux_bfid[:,3], aux_bfid[:,2], aux_bfid[:,0]))
  bf_reorder = np.eye(nb)[bf_reindex]
  # U and V are easily reordered
  # (this affects the matrices even if there was no rotation, so enforce rotation)
  do_rot = True
  U = U @ bf_reorder.T
  V = bf_reorder @ V

do_smth = do_rot or do_trans or do_exch or do_desym

# Process the input data and write the output file

# Known attributes and datasets that are generally safe to copy (need no transformation)
safe_to_copy = [
  # attributes
  'A2LEV',
  'CI_TYPE',
  'L2ACT',
  'MOLCAS_MODULE',
  'MOLCAS_VERSION',
  'NACTEL',
  'NCONF',
  'NELEC3',
  'NHOLE1',
  'NPRIM',
  'NROOTS',
  'NSTATE',
  'NSTATES',
  'ORBITAL_TYPE',
  'POTNUC',
  'RASSCF_ITERATIONS',
  'ROOT2STATE',
  'SPINMULT',
  'STATE_SPINMULT',
  # datasets
  'CI_VECTORS',
  'ENERGY',
  'H_EFF',
  'ORIGINAL_OVERLAPS',
  'ROOT_ENERGIES',
  'SFS_COEFFICIENTS',
  'SFS_ENERGIES',
  'STATE_WEIGHT',
  'STATE_PT2_ENERGIES',
  'STATE_REFWF_ENERGIES',
  'STATE_ROOTID',
]

# Attributes and datasets that cannot be transformed with desymmetrization 
# (e.g. state order is not known)
not_with_desym = [
  # attributes
  'NCONF',
  # datasets
  'CI_VECTORS',
  'STATE_ROOTID',
]

warning_item = []
warning_renum = False
with h5py.File(args['infile'], 'r') as fi, h5py.File(args['outfile'], 'w') as fo:

  fo.attrs.create('MORT_version', np.array(f'MORT v{version}', dtype='S'))

  module = fi.attrs['MOLCAS_MODULE'].decode()

  # Pick up the data needed for desymmetrization, which includes reordering MOs
  if do_desym:
    desym = np.reshape(fi['DESYM_MATRIX'], (nb, nb))
    nbas = fi.attrs['NBAS']
    mo_reindex = np.lexsort((fi['MO_ENERGIES'], -fi['MO_OCCUPATIONS'][:]))
    try:
      # In case of UHF
      mo_alpha_reindex = np.lexsort((fi['MO_ALPHA_ENERGIES'], -fi['MO_ALPHA_OCCUPATIONS'][:]))
      mo_beta_reindex = np.lexsort((fi['MO_BETA_ENERGIES'], -fi['MO_BETA_OCCUPATIONS'][:]))
    except KeyError:
      pass
    try:
      # In case of [CR]ASSCF, some data is on active orbitals only
      tpidx = fi['MO_TYPEINDICES'][:]
    except KeyError:
      pass
    else:
      actidx = np.isin(tpidx, [b'1', b'2', b'3']).nonzero()[0]
      nact = actidx.shape[0]
      act_reindex = np.empty(nact)
      for i in range(nact):
        act_reindex[i] = np.where(mo_reindex == actidx[i])[0][0]
      act_reindex = np.argsort(act_reindex)
      act_reorder = np.eye(nact)[act_reindex]
      # In CASPT2 density matrices are stored for all (non-frozen, non-deleted) orbitals
      if module == 'CASPT2':
        pt2idx = np.logical_not(np.isin(tpidx, [b'F', b'D'])).nonzero()[0]
        bins = [0]
        for i in nbas:
          bins.append(bins[-1]+i)
        pt2nbas = np.histogram(pt2idx, bins=bins)[0]
        # TODO: reindex/reorder MOs for CASPT2 (even if nothing happens by default)
    V = V @ desym.T
  else:
    nbas = [nb]

  for name in fi.attrs:

    if name in safe_to_copy:
      if do_desym and name in not_with_desym:
        continue
      fo.attrs[name] = fi.attrs[name]

    elif name == 'LSYM':
      if do_desym:
        fo.attrs[name] = 1
      else:
        fo.attrs[name] = fi.attrs[name]

    elif name == 'NSYM':
      if do_desym:
        fo.attrs[name] = 1
      else:
        fo.attrs[name] = fi.attrs[name]

    elif name == 'IRREP_LABELS':
      if do_desym:
        fo.attrs[name] = np.array(['a  '], dtype='S')
      else:
        fo.attrs[name] = fi.attrs[name]

    elif name == 'STATE_IRREPS':
      fo.attrs[name] = fi.attrs[name]
      if do_desym:
        fo.attrs[name][:] = 1

    elif name == 'NBAS':
      if do_desym:
        fo.attrs[name] = [np.sum(fi.attrs[name])]
      else:
        fo.attrs[name] = fi.attrs[name]

    elif name in ['NATOMS_UNIQUE', 'NATOMS_ALL']:
      if do_desym:
        name2 = 'NATOMS_UNIQUE'
        if name == name2:
          continue
        fo.attrs[name2] = fi.attrs[name]
      else:
        fo.attrs[name] = fi.attrs[name]

    elif args['all']:
      fo.attrs[name] = fi.attrs[name]
      warning_item.append(name)

  for name in fi:

    if name in safe_to_copy:
      if do_desym and name in not_with_desym:
        continue
      copy_dataset(fi, fo, name)

    elif name in ['DESYM_MATRIX']:
      continue

    # Coordinates are transformed with the input rotation
    elif name in ['CENTER_COORDINATES', 'DESYM_CENTER_COORDINATES']:
      name2 = name
      if name.startswith('DESYM_'):
        name2 = name[6:]
      elif do_desym:
        continue
      copy_dataset(fi, fo, name, name2)
      if do_rot or do_trans or do_exch:
        coor = fi[name]
        fo[name2][:] = reorder @ (coor @ R.T + T)

    # Multipole centers too, but they are not reordered
    elif name == 'MLTPL_ORIG':
      copy_dataset(fi, fo, name)
      if do_rot or do_trans:
        coor = fi[name]
        fo[name][:] = coor @ R.T + T

    # Other center data are reordered only
    elif name in ['CENTER_ATNUMS', 'CENTER_CHARGES', 'DESYM_CENTER_ATNUMS', 'DESYM_CENTER_CHARGES']:
      name2 = name
      if name.startswith('DESYM_'):
        name2 = name[6:]
      elif do_desym:
        continue
      copy_dataset(fi, fo, name, name2)
      if do_exch:
        data = fi[name]
        fo[name2][:] = data[:][reindex]

    # Special care with strings (can't just copy dataset because lengths may be different)
    elif name in ['CENTER_LABELS', 'DESYM_CENTER_LABELS']:
      name2 = name
      if name.startswith('DESYM_'):
        name2 = name[6:]
      elif do_desym:
        continue
      copy_dataset(fi, fo, name, name2, size=(nat,))
      if do_exch or do_desym:
        data = fi[name][:]
        # Find out if labels need renumbering
        any_num = False
        do_renum = True
        for i,lab in enumerate(fi[name2]):
          match = re.match(r'[A-Za-z]{1,2}(\d+)\s*', lab.decode())
          # Need renumbering if all labels have a number and all match the index
          if match:
            any_num = True
            if int(match.group(1)) != i+1:
              do_renum = False
          else:
            do_renum = False
        if do_renum:
          for i,lab in enumerate(data[reindex]):
            data[reindex[i]] = re.sub(r'^([A-Za-z]{1,2})(\d+).{4}', r'\g<1>'+f'{i+1}    ', lab.decode())
        elif any_num:
          warning_renum = True
        fo[name2][:] = data[reindex]
      else:
        fo[name2][:] = fi[name][:]

    # Basis info must be reordered and reindexed
    elif name in ['BASIS_FUNCTION_IDS', 'DESYM_BASIS_FUNCTION_IDS']:
      name2 = name
      if name.startswith('DESYM_'):
        name2 = name[6:]
      elif do_desym:
        continue
      copy_dataset(fi, fo, name, name2)
      if do_exch or do_desym:
        bfid = fi[name][:][bf_reindex]
        for b in bfid:
          b[0] = reindex_inv[b[0]-1]+1
        fo[name2][:] = bfid

    # Primitive info is similar to basis, but it was not read before
    elif name in ['PRIMITIVES', 'PRIMITIVE_IDS']:
      # Just to make sure these are processed in the right order
      if name == 'PRIMITIVES':
        continue
      copy_dataset(fi, fo, name)
      name = 'PRIMITIVES'
      copy_dataset(fi, fo, name)
      if do_exch:
        name = 'PRIMITIVE_IDS'
        prid = fi[name][:]
        pr_reindex = []
        for i in reindex:
          pr_reindex.extend([j for j,p in enumerate(prid) if p[0] == i+1])
        for p in prid:
          p[0] = reindex_inv[p[0]-1]+1
        fo[name][:] = prid[pr_reindex]
        name = 'PRIMITIVES'
        prim = fi[name][:][pr_reindex]
        fo[name][:] = prim

    # MO coefficients are transformed with the U matrix
    elif name == 'MO_VECTORS':
      copy_dataset(fi, fo, name, size=(nb**2,))
      if do_rot or do_desym:
        mo = sblock_to_square(fi[name], nbas)
        if do_desym:
          mo = (mo @ desym)[mo_reindex]
        fo[name][:] = (mo @ U).flatten()
    elif name == 'MO_ALPHA_VECTORS':
      copy_dataset(fi, fo, name, size=(nb**2,))
      if do_rot or do_desym:
        mo = sblock_to_square(fi[name], nbas)
        if do_desym:
          mo = (mo @ desym)[mo_alpha_reindex]
        fo[name][:] = (mo @ U).flatten()
    elif name == 'MO_BETA_VECTORS':
      copy_dataset(fi, fo, name, size=(nb**2,))
      if do_rot or do_desym:
        mo = sblock_to_square(fi[name], nbas)
        if do_desym:
          mo = (mo @ desym)[mo_beta_reindex]
        fo[name][:] = (mo @ U).flatten()

    # Other MO data are just reordered if desymmetrized
    elif name in ['MO_ENERGIES', 'MO_OCCUPATIONS', 'MO_TYPEINDICES', 'SUPSYM_IRREP_INDICES']:
      copy_dataset(fi, fo, name)
      if do_desym:
        data = fi[name]
        fo[name][:] = data[:][mo_reindex]
    elif name in ['MO_ALPHA_ENERGIES', 'MO_ALPHA_OCCUPATIONS', 'MO_ALPHA_TYPEINDICES']:
      copy_dataset(fi, fo, name)
      if do_desym:
        data = fi[name]
        fo[name][:] = data[:][mo_alpha_reindex]
    elif name in ['MO_BETA_ENERGIES', 'MO_BETA_OCCUPATIONS', 'MO_BETA_TYPEINDICES']:
      copy_dataset(fi, fo, name)
      if do_desym:
        data = fi[name]
        fo[name][:] = data[:][mo_beta_reindex]

    # AO matrices are transformed with the V matrix
    elif name in ['AO_OVERLAP_MATRIX', 'AO_FOCKINT_MATRIX']:
      copy_dataset(fi, fo, name, size=(nb**2,))
      if do_rot or do_desym:
        mtrx = sblock_to_square(fi[name], nbas)
        fo[name][:] = (V @ mtrx @ V.T).flatten()

    # Multipole AO matrices are further transformed among their components
    elif name in ['AO_MLTPL_X', 'AO_MLTPL_Y', 'AO_MLTPL_Z']:
      comp = ['X', 'Y', 'Z']
      # Do all work only when the first component is found
      if name.endswith(('_'+comp[1], '_'+comp[2])):
        continue
      ao_mltpl1 = np.empty((len(comp), nb, nb))
      for i,lab in enumerate(comp):
        name = 'AO_MLTPL_'+lab
        copy_dataset(fi, fo, name)
        ao_mltpl1[i,:,:] = fi[name]
      if do_rot:
        # Transform each component
        for i in range(len(comp)):
          ao_mltpl1[i,:,:] = V @ ao_mltpl1[i,:,:] @ V.T 
        # Combine components
        ao_mltpl1 = np.tensordot(R, ao_mltpl1, axes=1)
        for i,lab in enumerate(comp):
          name = 'AO_MLTPL_'+lab
          fo[name][:] = ao_mltpl1[i,:,:]
    elif name in ['AO_MLTPL_XX', 'AO_MLTPL_XY', 'AO_MLTPL_XZ', 'AO_MLTPL_YY', 'AO_MLTPL_YZ', 'AO_MLTPL_ZZ']:
      comp = ['XX', 'XY', 'XZ', 'YY', 'YZ', 'ZZ']
      # Do all work only when the first component is found
      if name.endswith(('_'+comp[1], '_'+comp[2], '_'+comp[3], '_'+comp[4], '_'+comp[5])):
        continue
      ao_mltpl2 = np.empty((len(comp), nb, nb))
      for i,lab in enumerate(comp):
        name = 'AO_MLTPL_'+lab
        copy_dataset(fi, fo, name)
        ao_mltpl2[i,:,:] = fi[name]
      if do_rot:
        # Transform each component
        for i in range(len(comp)):
          ao_mltpl2[i,:,:] = V @ ao_mltpl2[i,:,:] @ V.T 
        # Quadrupoles transform like a Cartesian d shell (the inverse)
        if -2 in RotM:
          QRot = np.linalg.inv(RotM[-2])
        else:
          QRot = np.linalg.inv(cartesian_shell_rotation(2, R))
        # Combine components
        ao_mltpl2 = np.tensordot(QRot, ao_mltpl2, axes=1)
        for i,lab in enumerate(comp):
          name = 'AO_MLTPL_'+lab
          fo[name][:] = ao_mltpl2[i,:,:]

    # Density matrices may need to be reordered if desymmetrizing
    elif name in ['DENSITY_MATRIX', 'SPINDENSITY_MATRIX', 'TRANSITION_DENSITY_MATRIX', 'TRANSITION_SPIN_DENSITY_MATRIX']:
      copy_dataset(fi, fo, name)
      if  do_desym:
        if module == 'CASPT2':
          # In CASPT2, DM is in triangular storage
          dmsize = list(fo[name].shape)
          del fo[name]
          n = int(np.sum(pt2nbas))
          dmsize[1] = n*(n+1)//2
          copy_dataset(fi, fo, name, size=dmsize)
          for i in range(dmsize[0]):
            dm = sblock_to_square(fi[name][i,:], pt2nbas, triangular=True)
            fo[name][i,:] = dm[np.tril_indices(n)]
        else:
          dm = fi[name][:]
          for i in range(dm.shape[0]):
            dm[i,:,:] = act_reorder @ dm[i,:,:] @ act_reorder.T
            fo[name][:] = dm

    elif args['all']:
      copy_dataset(fi, fo, name)
      warning_item.append(name)

# In case something unsafe was done
if warning_item and do_smth:
  print('WARNING: The following attributes and/or datasets are not recognized, and were copied verbatim.')
  print('         If they would need transformation, they are probably WRONG!')
  print()
  for name in warning_item:
    print(name)
if warning_renum:
  print('WARNING: Atoms were reordered, but labels are not renumbered')



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Oskar Weser                                      *
#***********************************************************************

__version__ = '0.0.0'

from analyze_molcas.spin_orbfile import SpinOrbs
from analyze_molcas.spatial_orbfile import SpatialOrbs



# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2017-2019, Ignacio Fdez. Galván                        *
#***********************************************************************

from __future__ import (unicode_literals, division, absolute_import, print_function)

from os.path import basename
from re import compile as re_compile
from molcas_aux import *

def check_test(infofile, checkfile, count):
  block = re_compile(r'#>>\s*(\d+)')
  reference = re_compile(r'#>\s*(.*)="(.*)"\/(.*)')
  rc = '_RC_ALL_IS_WELL_'
  last = True

  # Read actual values from the calculation
  vals = []
  try:
    with utf8_open(infofile) as i_file:
      for line in i_file:
        if (line.lstrip()[0] == '*'):
          continue
        match = reference.match(line)
        if match:
          try:
            vals.append({'lab': match.group(1),
                         'val': float(match.group(2)),
                         'tol': int(match.group(3)),
                         'raw': line.strip()})
          except:
            print('*** Error in {0}: {1}'.format(basename(infofile), line.rstrip()))
            return '_RC_CHECK_ERROR_'
  except:
    pass

  # If generating check file, write the values
  if (get_utf8('MOLCAS_TEST', default='').upper().startswith('GENE')):
    try:
      maxcount = int(get_utf8('MOLCAS_MAX_CHECKS', default=''))
    except:
      maxcount = 500
    if (count > maxcount):
      print('*** Too many program calls in test file')
      return '_RC_CHECK_ERROR_'
    print('\nGenerating check file')
    if (count == 1):
      with utf8_open(checkfile, 'w') as c_file:
        c_file.write('* This file is autogenerated:\n')
        for line in get_utf8('MOLCAS_INFO', default='').split('!'):
          prefix = '*'
          if (len(line) > 0):
            prefix += ' '
          c_file.write('{0}{1}\n'.format(prefix, line))
    with utf8_open(checkfile, 'a') as c_file:
      c_file.write('#>> {0:3}\n'.format(count))
      for item in vals:
        c_file.write('{0}\n'.format(item['raw']))

  # If checking, read the reference values
  elif (get_utf8('MOLCAS_TEST', default='').upper().startswith('CHECK')):
    start = False
    refs = []
    try:
      with utf8_open(checkfile) as c_file:
        for line in c_file:
          if (line.lstrip()[0] == '*'):
            continue
          match = block.match(line)
          if (match):
            if (start):
              last = False
              break
            if (int(match.group(1)) == count):
              start = True
            continue
          if (start):
            try:
              match = reference.match(line)
              refs.append({'lab': match.group(1),
                           'val': float(match.group(2)),
                           'tol': int(match.group(3))})
            except:
              print('*** Error in {0}: {1}'.format(basename(checkfile), line.rstrip()))
              return '_RC_CHECK_ERROR_'
    except:
      pass

    # Compare actual and reference values, and print summary
    start = False
    print('\nChecking results:')
    factor = int(get_utf8('MOLCAS_THR', default='0'))
    passcheck = int(get_utf8('MOLCAS_PASSCHECK', default='0'))
    failrc = '_RC_ALL_IS_WELL_' if (passcheck) else '_RC_CHECK_ERROR_'
    fuzzy = get_utf8('MOLCAS_CHECK_FUZZY', default='').upper()
    fuzzy = ((fuzzy == 'YES') or (fuzzy == 'ON'))
    if (factor > 0):
      print('\n*** Tolerance increased by a factor of {0}'.format(10**factor))
    else:
      factor = 0
    if (passcheck):
      print('\n*** This check is informative only, nothing fails')
    fmt_head = '\n{0:^30} {1:^20} {2:^20} {3:^9} {4:^9}'.format('Label','Value','Reference','Error','Tolerance')
    # maximum exponent is 3 digits long, this makes at most 7 extra characters (exponent, decimal point, "e" and two signs)
    # but error and tolerance are absolute values
    fmt_num = '{0:<30} {1:20.13g} {2:20.13g} {3:9.3e} {4:9.3e} {5}'
    fmt_rule = '-'*92
    j = -1
    failtol = False
    for i in range(len(refs)):
      # Find the corresponding item in the actual values
      j += 1
      extra = []
      while ((j < len(vals)) and (refs[i]['lab'] != vals[j]['lab'])):
        extra.append(vals[j]['lab'])
        j += 1
        if (not fuzzy):
          rc = failrc
      if (j >= len(vals)):
        print('*** Label not found: {0}'.format(refs[i]['lab']))
        rc = failrc
        break
      else:
        for item in extra:
          print('*** Extra label: {0}'.format(item))
      # Check the difference against the tolerance (with some dirty tricks)
      if (vals[j]['tol'] != refs[i]['tol']):
        failtol = True
      dif = abs(vals[j]['val'] - refs[i]['val'])
      tol = -refs[i]['tol']
      if (tol < 0):
        tol = 10**(tol+factor)
      else:
        tol = tol*10**(factor)
      if (dif > 0.0):
        if (not start):
          print(fmt_head)
          print(fmt_rule)
          start = True
        if (dif > tol):
          if ((refs[i]['lab'] == 'POTNUC') and (dif < tol*10)):
            tag = 'Skipped'
          elif (refs[i]['lab'].endswith('ITER') and (dif < tol+5)):
            tag = 'Skipped'
          else:
            tag = 'Failed'
            rc = failrc
        else:
          tag = ''
        print(fmt_num.format(refs[i]['lab'], vals[j]['val'], refs[i]['val'], dif, tol, tag))
    # Additional values at the end
    j += 1
    while (j < len(vals)):
      print('*** Extra label: {0}'.format(vals[j]['lab']))
      j += 1
      if (not fuzzy):
        rc = failrc
    if (failtol):
      print('*** Tolerances do not match')
      rc = failrc
    if (start):
      print(fmt_rule)
    elif (rc == '_RC_ALL_IS_WELL_'):
      print('All values match the reference')

  return rc, last



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2018, Alessio Valentini                                *
#               2018, Luis Manuel Frutos                               *
#               2021,2022, Jonathan Richard Church                     *
#               2021,2022, Igor Schapiro                               *
#***********************************************************************

import numpy as np
import random
import argparse
import os
import sys
import math

def printDict(dictionary):
    '''
    pretty printer for dictionary
    dictionary :: Dictionary
    '''
    for x in dictionary:
        print('{} -> {}'.format(x,dictionary[x]))

def test_initial_things():
    '''
    This function returns a dictionary with test inputs instead of reading them
    dictio :: Dictionary
    These values can be used as a test. I need to set the seed and the results.
    '''
    dictio = {}
    dictio['amu_to_kg'] = 1.66053892173E-27
    dictio['joule_to_kcal'] = 1.439326E+20
    dictio['bohr_to_m'] = 5.2917724900001E-11
    dictio['degrN'] = 3
    dictio['atomN'] = 3
    dictio['T'] = 298.15
    dictio['hbar'] = 1.0545718E-34
    dictio['kb'] = 1.38064852E-23
    dictio['beta'] = 1 / (dictio['T'] * 1.38064852E-23)
    dictio['c'] = 299792458.00
    dictio['cm_to_SI']=dictio['c'] * 100
    dictio['NCMatx'] = np.array([
        [[0.00,0.00,0.07],[0.00,-0.43,-0.56],[0.00,0.43,-0.56]],
        [[0.00,0.00,0.05],[0.00,0.58,-0.40],[0.00,-0.58,-0.40]],
        [[0.00,0.07,0.00],[0.00,-0.55,0.44],[0.00,-0.55,-0.44]]])
    dictio['geom'] = np.array([
             [0.000000, 0.000000, 0.119736],
             [0.000000, 0.761603,-0.478944],
             [0.000000,-0.761603,-0.478944]])
    dictio['RedMass'] = np.array([1.0825,1.0453,1.0810])
    dictio['AtMass'] = np.array([15.9994,1.00794,1.00794])
    dictio['freq'] = np.array([1713.0153, 3727.3731,3849.4717])
    dictio['debug'] = False
    dictio['atomT'] = ['O','H','H']
    return (dictio)

def gaus_dist(sigma):
    '''
    Generate a Gaussian distribution given a sigma
    sigma :: Double
    '''
    u1 = random.uniform(0, 1)
    u2 = random.uniform(0, 1)
    #u1 = 0.5
    #u2 = 0.8
    z = np.sqrt(-2*np.log(u1))*np.sin(2*np.pi*u2)
    return z*sigma

def normal_mode(dictio,label,method):
    '''
    Main driver for initial condition with normal mode sampling. Takes as input a dictionary of inputs.
    dictio :: Dictionary <- inputs
    label :: String <- project name
    in this routine the Boltzmann distribution is calculated for a single initial condition

    degrN :: Int <- number of degrees of freedom
    atomN :: Int <- number of atoms

    some dimensions
    NCMatx :: (degrN, atomN, 3)
    Pcart,vcart :: (atomN,3)

    '''

    # dictionary unpacking
    c = dictio['c']
    beta = dictio['beta']
    freq = dictio['freq']
    geom = dictio['geom']
    atomN = dictio['atomN']
    NCMatx = dictio['NCMatx']
    AtMass = dictio['AtMass']
    RedMass = dictio['RedMass']
    degrN = dictio['degrN']
    debug = dictio['debug']
    atomT = dictio['atomT']
    kb = dictio['kb']
    hbar = dictio['hbar']
    cm_to_SI = dictio['cm_to_SI']
    amu_to_kg = dictio['amu_to_kg']
    joule_to_kcal = dictio['joule_to_kcal']
    bohr_to_m = dictio['bohr_to_m']
    T=dictio['T']
    ##Generate mass matrix from atoms
    mmatrix=mass_matrix(AtMass, atomN)*amu_to_kg
    ##Convert Coordinates to meters for calculations
    xyz_save=geom*bohr_to_m
    xyz_old=np.reshape(xyz_save, (1, 3*atomN))
    ##reshape eigenvector matricices
    NCMatx=np.reshape(NCMatx, (degrN, 3*atomN))
    ##create blank matricies for soon to be created velocities and displacements
    coord_samp=np.zeros((1, 3*atomN), dtype=float)
    ##need to add a test for linearity
    coord_samp_save=np.zeros((degrN, 3*atomN), dtype=float)
    velocity_samp=np.zeros((1, 3*atomN), dtype=float)
    ##This is a test to make sure there were no frequencies included in the hessian calculation from numerical error
    ##Count over normal modes with frequencies greater than 10cm-1 (assumed everything below that is error from the hessian calculation)
    modes=len(freq[freq > 10])
    ##If the number of modes is equal to degrN then proceed assuming that there were no numerical errors, otherwise include all modes but only use those greater than 10cm-1
    if (modes != degrN):
        j=int((3*atomN)-modes)
        COM_modes=j
    else:
        j=0
        COM_modes=0
    ##set variables and begin loop for normal mode analysis or wigner distribution
    Etot=0.0
    i=0
    while (j < degrN):
        ##Convert frequency to SI units
        freqSI=float(freq[j])*cm_to_SI*np.pi*2.0
        ##Normal Mode Sampling for vibrational ground state using a classical boltzmann distribution of energies for each normal mode
        if (method==2):
            ##Generate random number
            rand1=random.uniform(0, 1)
            ##Generate Initial guess of Ei from the cumulative distribution function of normal mode i at temperature T
            Ei=kb*T*(1-rand1)
            Etot=Ei+Etot
            ##Convert Frequencies to joules and calculate Amplitude
            A=np.power((2.0*Ei),0.5)/freqSI
            #Calculate normal coordinate and normal momentum in SI units
            x=A*np.cos(2.0*np.pi*rand1)
            v=-freqSI*A*math.sin(2.0*np.pi*rand1)
        ##Wigner sampling for ground vibrational state
        if (method==3):
            sample=0
            while (sample==0):
                rand1=random.uniform(-1, 1)*np.sqrt(hbar/freqSI)
                rand2=random.uniform(-1, 1)*np.sqrt(hbar*freqSI)
                rand3=random.uniform(0, 1)
                Ei=0.5*(np.power(freqSI*rand1,2)+np.power(rand2,2))
                probability=1.0/(np.pi)*np.exp(-(2.0*Ei)/(hbar*freqSI))
                if (probability > rand3/np.pi):
                    sample=sample+1
                    x=rand1
                    v=rand2
                    Etot=Etot+Ei
        ##Thermal Wigner sampling
        if (method==4):
            sample=0
            while (sample==0):
                alpha=np.tanh(hbar*freqSI/(2.0*kb*T))
                rand1=random.uniform(-1, 1)*np.sqrt(hbar/(freqSI*alpha))
                rand2=random.uniform(-1, 1)*np.sqrt(hbar*freqSI*(1.0/alpha))
                rand3=random.uniform(0, 1)
                Ei=0.5*(np.power(freqSI*rand1,2)+np.power(rand2,2))
                probability=alpha/(np.pi)*np.exp(-2.0*alpha*(np.power(freqSI*rand1,2)+np.power(rand2,2))/(hbar*freqSI))
                if (probability > rand3/(np.pi/alpha)):
                    sample=sample+1
                    x=rand1
                    v=rand2
                    Etot=Etot+Ei
        ##Generate displacements and velocities based on sampling method
        coord_samp=coord_samp+x*NCMatx[j, :]/np.sqrt(mmatrix)
        coord_samp_save[i, :]=x*NCMatx[j, :]/np.sqrt(mmatrix)
        velocity_samp=velocity_samp+v*NCMatx[j, :]/np.sqrt(mmatrix)
        j=j+1
        i=i+1
    ##Add displacement to optimized coordinates and generate cartesian velocities
    vel_new=velocity_samp
    xyz_new=xyz_old+coord_samp
    vel_reshape=np.reshape(vel_new, (atomN, 3))
    xyz_reshape=np.reshape(xyz_new, (atomN, 3))
    ##Check initial total energy based on harmonic oscillator hamiltonian
    E=0.0
    j=COM_modes
    KE=np.power(mmatrix[:]*np.reshape(vel_reshape, (1, 3*atomN)), 2.0)/(2.0*mmatrix[:])
    KE=np.sum(KE)
    while (j < degrN):
        Ei=np.power(float(freq[j])*cm_to_SI*2.0*np.pi, 2.0)*np.power(np.power(mmatrix[:], 0.5)*coord_samp_save[j-COM_modes, :], 2.0)/2.0
        E=E+np.sum(Ei)
        j=j+1
    E=E+KE  
    #This can be uncommented for testing print("initial E, KE, Etot", E-KE, " ", KE, " " , Etot, '\n')
    ##Begin to removal any supurious COM translation or rotation in the molecule and then adjusting the velocities and displacements to conserve energy
    accept=0.0
    Mass=AtMass*amu_to_kg
    while (accept==0):
        ##First calculate the cartesian coordinates with the COM at origin
        com=CenterOfMass(Mass, vel_reshape, atomN)
        com_xyz=CenterOfMass(Mass, xyz_reshape, atomN)
        xyz_reshape_COM=xyz_reshape
        xyz_reshape_COM[:, 0] -= com_xyz[0]
        xyz_reshape_COM[:, 1] -= com_xyz[1]
        xyz_reshape_COM[:, 2] -= com_xyz[2]
        ##Next generate angular momentum, inverse moment of interia matrix, and angular velocity
        Ltot=angular_mo(Mass, xyz_reshape_COM, vel_reshape, atomN)
        invI=inertia(xyz_reshape_COM, Mass, atomN)
        ang_vel=np.dot(invI, Ltot)
        xyz_reshaped=np.reshape(xyz_reshape_COM, (atomN, 3))
        j=0
        while (j < atomN):
            ##Remove spurious rotational velocity
            x=xyz_reshaped[j,0]
            y=xyz_reshaped[j,1]
            z=xyz_reshaped[j,2]
            xyz_corr=np.array([x, y, z])
            ang_corr=angular_vel(Mass, xyz_corr, ang_vel, atomN)
            vel_reshape[j, 0] -= ang_corr[0]
            vel_reshape[j, 1] -= ang_corr[1]
            vel_reshape[j, 2] -= ang_corr[2]
            j=j+1
        ##Remove spurious COM velocity
        com=CenterOfMass(Mass, vel_reshape, atomN)
        vel_reshape[:, 0] -= com[0]
        vel_reshape[:, 1] -= com[1]
        vel_reshape[:, 2] -= com[2]
        ##Calculate harmonic energy and compare to original value
        E=0.0
        j=COM_modes
        KE=np.power(mmatrix[:]*np.reshape(vel_reshape, (1, 3*atomN)), 2.0)/(2.0*mmatrix[:])
        KE=np.sum(KE)
        while (j < degrN):
            Ei=np.power(float(freq[j])*cm_to_SI*2.0*np.pi, 2.0)*np.power(np.power(mmatrix[:], 0.5)*coord_samp_save[j-COM_modes, :], 2.0)/2.0
            E=E+np.sum(Ei)
            j=j+1
        E=E+KE
        #This can be uncommented for testing print("after error removal E, KE, Etot", E-KE, " ", KE, " " , Etot)
        ##If the  value differs by less than 1 percent accept and move to next conndition, otherwise scale the displacements and velocities and try again
        if (abs(E-Etot)/Etot*100 < 1):
            accept=accept+1
            vel_final=vel_reshape
            coord_final=xyz_reshape
        else :
            vel_reshape=vel_reshape*np.power(Etot/E, 0.5)
            xyz_reshape=np.reshape(xyz_old+(coord_samp)*np.power(Etot/E, 0.5), (atomN, 3))
            coord_samp_save=(coord_samp_save)*np.power(Etot/E, 0.5)
    ##Prepare for final coordinate and velocity file writing
    xyz_final=np.reshape(coord_final, (atomN, 3))
    vel_final=np.reshape(vel_final, (atomN, 3))
    ##Convert coordinates to angstroms, velocities to bohr/au
    xyz_final=xyz_final*1.0E10
    vel_final=vel_final*4.57102844e-7
    ##Create final file name
    geomName = label + '.xyz'
    veloName = label + '.velocity.xyz'
    #stringOUT = '\n{}\n{}\n{}\n{}'
    #print(stringOUT.format(veloName, Qcart, geomName, newGeom))
    np.savetxt(veloName,vel_final,fmt='%1.6f')
    # add the atom name in the matrix
    atom_type = np.array(atomT)[:, np.newaxis]
    atom_t_and_geom = np.hstack((atom_type,xyz_final))
    np.savetxt(geomName,atom_t_and_geom,header='{}\n'.format(atomN),comments='',fmt='%s %s %s %s')

def generate_one_boltz(dictio,label):
    '''
    Main driver for initial condition. Takes as input a dictionary of inputs.
    dictio :: Dictionary <- inputs
    label :: String <- project name
    in this routine the Boltzmann distribution is calculated for a single initial condition

    degrN :: Int <- number of degrees of freedom
    atomN :: Int <- number of atoms

    some dimensions
    NCMatx :: (degrN, atomN, 3)
    Pcart,vcart :: (atomN,3)

    '''

    # dictionary unpacking
    c = dictio['c']
    beta = dictio['beta']
    freq = dictio['freq']
    geom = dictio['geom']
    atomN = dictio['atomN']
    NCMatx = dictio['NCMatx']
    AtMass = dictio['AtMass']
    RedMass = dictio['RedMass']
    degrN = dictio['degrN']
    debug = dictio['debug']
    atomT = dictio['atomT']
    kb = dictio['kb']
    hbar = dictio['hbar']
    cm_to_SI = dictio['cm_to_SI']
    amu_to_kg = dictio['amu_to_kg']
    joule_to_kcal = dictio['joule_to_kcal']
    bohr_to_m = dictio['bohr_to_m']
    # DEBUG TIME
    if debug:
        printDict(dictio)
        import pickle
        with open('debug_dictionary_file.pkl', 'wb') as f:
            pickle.dump(dictio, f, pickle.HIGHEST_PROTOCOL)
    # this is lambda :: (degrN)
    lamb = 4 * (np.pi**2) * (freq*100)**2 * c**2
    sigmaP = np.sqrt(RedMass * 1.660537E-27 / beta)
    # change of dimensions
    Pint = (np.array([ gaus_dist(x) for x in sigmaP ]))*6.02214086E21
    # I want to multiply on first axis, both for Pcart and vcart
    # this is why I create and tile the new pint_broadcasted vector
    # it is ugly, there must be a broadcast numpy rule that works better and more efficient
    # but since the degrees of freedom are never gonna be 30.000.000, this loop here is
    # still efficient enough.
    pint_broadcasted = np.empty((degrN,atomN,3))
    for i in range(degrN):
        pint_broadcasted[i] = np.ones((atomN,3)) * Pint[i]
    to_be_summed = NCMatx * pint_broadcasted
    Pcart = np.sum(to_be_summed, axis=0)
    # same multiplication on first axis as with pint
    AtMass_broadcasted = np.empty((atomN,3))
    for i in range(atomN):
        AtMass_broadcasted[i] = np.ones(3) * AtMass[i]
    vcart = Pcart / AtMass_broadcasted
    sigma_q = 1/np.sqrt(beta*lamb)
    # change of dimensions
    Qint = (np.array([ gaus_dist(x) for x in sigma_q ]))*2.4540051E23
    # same multiplication on first axis as with pint and AtMass
    qint_broadcasted = np.empty((degrN,atomN,3))
    for i in range(degrN):
        qint_broadcasted[i] = np.ones((atomN,3)) * Qint[i]
    to_be_summed = NCMatx * qint_broadcasted
    Qcart_temp = np.sum(to_be_summed, axis=0)
    atmass_squared = np.sqrt(AtMass_broadcasted)
    Qcart = Qcart_temp/atmass_squared
    # Pcart is the displacement, added to the main geometry
    # Transformed into ANGSTROM !!
    # in Jan 2019, Dynamix takes geometries in angstrom but velocities in bohr
    newGeom = (Pcart + geom) * 0.529177249
    geomName = label + '.xyz'
    veloName = label + '.velocity.xyz'
    #stringOUT = '\n{}\n{}\n{}\n{}'
    #print(stringOUT.format(veloName, Qcart, geomName, newGeom))
    np.savetxt(veloName,Qcart,fmt='%1.6f')
    # add the atom name in the matrix
    atom_type = np.array(atomT)[:, np.newaxis]
    atom_t_and_geom = np.hstack((atom_type,newGeom))
    np.savetxt(geomName,atom_t_and_geom,header='{}\n'.format(atomN),comments='',fmt='%s %s %s %s')

def parseCL():
    d = '''
This tools is intended to be used as a support to launch molecular dynamics'

usage example:

python3 $MOLCAS/Tools/dynamixtools/dynamixtools.py -t 273 -c 100 -m 1 -i ${Project}.freq.molden

'''
    parser = argparse.ArgumentParser(description=d, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("-s", "--seed",
                        dest="seed",
                        required=False,
                        type=int,
                        help="indicate the SEED to use for the generation of randoms")
    parser.add_argument("-l", "--label",
                        dest="label",
                        required=False,
                        type=str,
                        help="label for your project (default is \"geom\")")
    parser.add_argument("-i", "--input",
                        dest="i",
                        required=False,
                        type=str,
                        help="path of the frequency h5 or molden file")
    parser.add_argument("-c", "--condition",
                        dest="condition",
                        required=False,
                        type=int,
                        help="number of initial conditions (default 1)")
    parser.add_argument("-t", "--temperature",
                        dest="temp",
                        required=False,
                        type=float,
                        help="temperature in kelvin for the initial conditions")
    parser.add_argument("-v", "--verbose",
                        dest="debug",
                        required=False,
                        action='store_true',
                        help="more verbose output")
    parser.add_argument("-T", "--TEST",
                        dest="test",
                        required=False,
                        action='store_true',
                        help="keyword use to test the routines")
    parser.add_argument("-D", "--DIGIT",
                        dest="digits",
                        required=False,
                        action='store_true',
                        help="keyword to suppress the counter in the filename (needed for debug)")
    parser.add_argument("-m", "--method",
                        dest="method",
                        required=False,
                        type=int,
                        help=('''\
Keyword to specify the sampling method:
1 Initial conditions based on the molecular vibrational frequencies and energies sampled from a Boltzmann distribution (Default).
2 Thermal normal mode sampling where the cumulitative distribution function for a classical boltzmann distribution at temperature T is used to approximate the energy of each mode.
3 Wigner distribution for the ground vibrational state, n=0.
4 Thermal Wigner distribution for temperature T based on the analytical solution for a canonical ensemble of harmonic oscillators.'''))
    args = parser.parse_args()
    return args

def parseMoldenFreq(fn):
    '''
    This function is the parser for molden freq files.
    '''
    inp = {}
    with open(fn) as f:
        first = f.readline()
        # parse n_freq
        second = f.readline()
        if second.strip() == '[N_FREQ]':
            degrN = int(f.readline().strip())
            inp['degrN'] = degrN
        else:
            sys.exit('This molden format is not recognized N_FREQ')
        # parse frequencies (they should be second)
        freqLabel = f.readline()
        if freqLabel.strip() == '[FREQ]':
            freq = np.empty(degrN)
            for i in range(degrN):
                freq_this_line = f.readline()
                freq[i] = float(freq_this_line.strip())
            inp['freq'] = freq
        else:
            sys.exit('This molden format is not recognized FREQ')
        # parse INT (they should be third)
        intLabel = f.readline()
        if intLabel.strip() == '[INT]':
            intL = np.empty(degrN)
            for i in range(degrN):
                intL_this_line = f.readline()
                intL[i] = float(intL_this_line.strip())
            inp['intL'] = intL
        else:
            sys.exit('This molden format is not recognized INT')
        # parse NATOM
        natomLabel = f.readline()
        if natomLabel.strip() == '[NATOM]':
            atomN = int(f.readline().strip())
            inp['atomN'] = atomN
        else:
            sys.exit('This molden format is not recognized NATOM')
        # parse FR-COORD (they should be fifth field)
        coordLabel = f.readline()
        if coordLabel.strip() == '[FR-COORD]':
            atomT = []
            coordL = np.empty((atomN,3))
            for i in range(atomN):
                geom_this_line = f.readline()
                # H 1.9 1.3 -4.5
                lol = filter(None, geom_this_line.strip('\n').split(' '))
                a,x,y,z = list(lol)
                atomT.append(a)
                coordL[i] = [float(x),float(y),float(z)]
            inp['atomT'] = atomT
            inp['geom'] = coordL
        else:
            sys.exit('This molden format is not recognized FR-COORD')
        # parse FR-NORM-COORD (they should be sixth field)
        norm_coord_Label = f.readline()
        if norm_coord_Label.strip() == '[FR-NORM-COORD]':
            NCMatx = np.empty((degrN,atomN,3))
            for j in range(degrN):
                vibration_line = f.readline()
                for i in range(atomN):
                    geom_this_line = f.readline()
                    evecs = filter(None, geom_this_line.strip('\n').split(' '))
                    x,y,z = list(evecs)
                    NCMatx[j,i] = [float(x),float(y),float(z)]
            inp['NCMatx'] = NCMatx
        else:
            sys.exit('This molden format is not recognized FR-NORM-COORD')
        # parse RMASS (they should be third)
        rmass_Label = f.readline()
        if rmass_Label.strip() == '[RMASS]':
            rmassL = np.empty(degrN)
            for i in range(degrN):
                rmassL_this_line = f.readline()
                rmassL[i] = float(rmassL_this_line.strip())
            inp['RedMass'] = rmassL
        else:
            sys.exit('This molden format is not recognized RMASS')
    # molden file finished, passing the dictionary back
    return(inp)

def parseh5Freq(fn):
    inp = {}
    sys.exit('This feature is still not available')
    return(inp)

def massOf(elem):
    '''
    You get the mass of an element from the label string
    elem :: String
    '''
    dictMass = {'X': 0, 'Ac': 227.028, 'Al': 26.981539, 'Am': 243, 'Sb': 121.757, 'Ar':
            39.948, 'As': 74.92159, 'At': 210, 'Ba': 137.327, 'Bk': 247, 'Be':
            9.012182, 'Bi': 208.98037, 'Bh': 262, 'B': 10.811, 'Br': 79.904,
            'Cd': 112.411, 'Ca': 40.078, 'Cf': 251, 'C': 12.011, 'Ce': 140.115,
            'Cs': 132.90543, 'Cl': 35.4527, 'Cr': 51.9961, 'Co': 58.9332, 'Cu':
            63.546, 'Cm': 247, 'Db': 262, 'Dy': 162.5, 'Es': 252, 'Er': 167.26,
            'Eu': 151.965, 'Fm': 257, 'F': 18.9984032, 'Fr': 223, 'Gd': 157.25,
            'Ga': 69.723, 'Ge': 72.61, 'Au': 196.96654, 'Hf': 178.49, 'Hs':
            265, 'He': 4.002602, 'Ho': 164.93032, 'H': 1.00794, 'In': 114.82,
            'I': 126.90447, 'Ir': 192.22, 'Fe': 55.847, 'Kr': 83.8, 'La':
            138.9055, 'Lr': 262, 'Pb': 207.2, 'Li': 6.941, 'Lu': 174.967, 'Mg':
            24.305, 'Mn': 54.93805, 'Mt': 266, 'Md': 258, 'Hg': 200.59, 'Mo': 95.94,
            'Nd': 144.24, 'Ne': 20.1797, 'Np': 237.048, 'Ni': 58.6934, 'Nb': 92.90638,
            'N': 14.00674, 'No': 259, 'Os': 190.2, 'O': 15.9994, 'Pd': 106.42, 'P':
            30.973762, 'Pt': 195.08, 'Pu': 244, 'Po': 209, 'K': 39.0983, 'Pr':
            140.90765, 'Pm': 145, 'Pa': 231.0359, 'Ra': 226.025, 'Rn': 222,
            'Re': 186.207, 'Rh': 102.9055, 'Rb': 85.4678, 'Ru': 101.07, 'Rf':
            261, 'Sm': 150.36, 'Sc': 44.95591, 'Sg': 263, 'Se': 78.96, 'Si':
            28.0855, 'Ag': 107.8682, 'Na': 22.989768, 'Sr': 87.62, 'S': 32.066,
            'Ta': 180.9479, 'Tc': 98, 'Te': 127.6, 'Tb': 158.92534, 'Tl':
            204.3833, 'Th': 232.0381, 'Tm': 168.93421, 'Sn': 118.71, 'Ti':
            47.88, 'W': 183.85, 'U': 238.0289, 'V': 50.9415, 'Xe': 131.29,
            'Yb': 173.04, 'Y': 88.90585, 'Zn': 65.39, 'Zr': 91.224}
    return(dictMass[elem])

def atomic_masses(atomtype_list):
    '''
    this function takes an atomtype list and return a numpy array with
    atomic masses
    atomtype_list :: [String] <- ['H', 'H', 'O']
    '''
    at_mass = [ massOf(x) for x in atomtype_list ]
    return np.array(at_mass)

def mass_matrix(AtMass, atomN):
    mass=np.zeros((atomN, 1), dtype=float)
    mass_matrix=np.zeros((atomN, 3), dtype=float)
    i=0
    while (i < atomN):
        mass_matrix[i, 0]=AtMass[i]
        mass_matrix[i, 1]=AtMass[i]
        mass_matrix[i, 2]=AtMass[i]
        i=i+1
    mass_matrix=np.reshape(mass_matrix, (1, 3*atomN))
    return mass_matrix

def inertia (xyz,mass,atomN):
    ##I'm not in love with this function and if anyone can do it better please do.
    j=0
    lxx=0
    lyy=0
    lzz=0
    lxy=0
    lxz=0
    lyx=0
    lzx=0
    lyz=0
    lzy=0
    while (j < 3):
        i=0
        while (i < 3):
            h=0
            while (h < atomN):
                stringm=float(mass[h])
                stringx=float(xyz[h, 0])
                stringy=float(xyz[h, 1])
                stringz=float(xyz[h, 2])
                if (i==j) and (i==0) and (j==0):
                    lxx=lxx+stringm*(math.pow(stringy, 2)+math.pow(stringz, 2))
                elif (i==j) and (i==1) and (j==1):
                    lyy=lyy+stringm*(math.pow(stringx, 2)+math.pow(stringz, 2))
                elif (i==j) and (i==2) and (j==2):
                    lzz=lzz+stringm*(math.pow(stringx, 2)+math.pow(stringy, 2))
                elif (i!=j) and (j==0) and (i==1):
                    lxy=lxy+(-stringm*stringx*stringy)
                elif (i!=j) and (j==0) and (i==2):
                    lxz=lxz+(-stringm*stringx*stringz)
                elif (i!=j) and (j==1) and (i==0):
                    lyx=lyx+(-stringm*stringy*stringx)
                elif (i!=j) and (j==1) and (i==2):
                    lyz=lyz+(-stringm*stringy*stringz)
                elif (i!=j) and (j==2) and (i==0):
                    lzx=lzx+(-stringm*stringz*stringx)
                elif (i!=j) and (j==2) and (i==1):
                    lzy=lzy+(-stringm*stringz*stringy)
        
                h=h+1
            i=i+1
        j=j+1
    data1=np.array([[lxx, lxy, lxz], [lyx, lyy, lyz], [lzx, lzy, lzz]])
    invI=np.linalg.inv(data1)
    return invI

def CenterOfMass(mass, vel, atomN):
    h=0
    stringcomx=0
    stringcomy=0
    stringcomz=0
    stringmtot=np.sum(mass[:])
    while (h < atomN):
        stringcomx=stringcomx+float(vel[h, 0])*float(mass[h])
        stringcomy=stringcomy+float(vel[h, 1])*float(mass[h])
        stringcomz=stringcomz+float(vel[h, 2])*float(mass[h])
        h=h+1
    comx=stringcomx/stringmtot
    comy=stringcomy/stringmtot
    comz=stringcomz/stringmtot
    com=np.array([comx, comy, comz])
    return com

def angular_mo (mass, xyz, vel, atomN):
    Lxtot=0
    Lytot=0
    Lztot=0
    h=0
    xyz=np.reshape(xyz, (atomN, 3))
    while (h < atomN):
        px=float(vel[h, 0])*float(mass[h])
        py=float(vel[h, 1])*float(mass[h])
        pz=float(vel[h, 2])*float(mass[h])
        Lx=float(xyz[h, 1])*pz-float(xyz[h, 2])*py
        Ly=float(xyz[h, 2])*px-float(xyz[h, 0])*pz
        Lz=float(xyz[h, 0])*py-float(xyz[h, 1])*px
        Lxtot=Lxtot+Lx
        Lytot=Lytot+Ly
        Lztot=Lztot+Lz
        h=h+1
    Ltot=np.array([Lxtot, Lytot, Lztot])
    return Ltot

def angular_vel (mass, xyz, vel, atomN):
    px=float(vel[0])
    py=float(vel[1])
    pz=float(vel[2])
    wx=-float(xyz[1])*pz+float(xyz[2])*py
    wy=-float(xyz[2])*px+float(xyz[0])*pz
    wz=-float(xyz[0])*py+float(xyz[1])*px
    wtot=np.array([wx, wy, wz])
    return wtot

def main():
    print('')
    args = parseCL()

    if args.test:
        inputs = test_initial_things()
        seedI = 123456789
        random.seed(seedI)
        complete_label = 'test_bolz'
        generate_one_boltz(inputs,complete_label)
        print(inputs)
    else:
        # i is input file from command line
        if args.i:
            fn=args.i
        else:
            # I do not like this termination here, but I still have to figure out how
            # to properly do mutually exclusive argparse keywords.
            # I will keep this exit code here in the meanwhile...
            sys.exit('-i input freq file is a required keyword, --help for help')
        if args.seed:
            seedI = args.seed
            print('seed set to: {}'.format(seedI))
            random.seed(seedI)
        if args.label:
            label = args.label
            print('Project label set to: {}'.format(label))
        else:
            label = 'geom'
        if args.condition:
            number_of_ic = args.condition
        else:
            number_of_ic = 1
        if args.method:
            method = args.method
        else:
            method = 1
        # check if is is molden or h5
        name, ext = os.path.splitext(fn)
        if ext == '.molden':
            inputs = parseMoldenFreq(fn)
        elif ext == '.h5':
            inputs = parseh5Freq(fn)
        else:
            print('You must use freq.molden or .h5 files')
        if args.debug:
            inputs['debug'] = True
        else:
            inputs['debug'] = False
        if args.temp == None:
            print('\nSetting default temperature to 300 K. Use the -t keyword to set a custom temperature.\n')
            inputs['T'] = 300
        else:
            inputs['T'] = args.temp
        ##Conversion Factors and constants
        inputs['kb'] = 1.38064852E-23
        inputs['h'] = 6.62607004E-34
        inputs['hbar'] = 1.0545718E-34
        inputs['beta'] = 1 / (inputs['T'] * inputs['kb'])
        inputs['c'] = 299792458.00
        inputs['cm_to_SI'] = inputs['c'] * 100
        inputs['amu_to_kg'] = 1.66053892173E-27
        inputs['joule_to_kcal'] = 1.439326E+20
        inputs['bohr_to_m'] = 5.2917724900001E-11
        inputs['AtMass'] = atomic_masses(inputs['atomT'])
        for counter in range(number_of_ic):
            if args.digits:
                complete_label = '{}'.format(label)
            else:
                complete_label = '{}{:04}'.format(label,counter)
            if (method==1):
                    generate_one_boltz(inputs,complete_label)
            elif (method==2 or method==3 or method==4):
                    normal_mode(inputs,complete_label,method)
        print('\nThis routine generates geometries in angstrom and velocities in bohr (the format that Molcas requires for a Semiclassical Molecular Dynamics\n')

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2022, Ignacio Fdez. Galván                             *
#***********************************************************************

import sphinx

# Patch to work around Sphinx bug #9529
if (3, 5, 0, '', 0) <= sphinx.version_info < (4, 5, 0, '', 0):
  from sphinx.writers.latex import CR
  from docutils import nodes
  from docutils.nodes import Element
  from typing import cast
  visit_start_of_file_orig = sphinx.writers.latex.LaTeXTranslator.visit_start_of_file
  visit_footnote_orig = sphinx.writers.latex.LaTeXTranslator.visit_footnote
  def visit_start_of_file_patched(self, node: Element) -> None:
    visit_start_of_file_orig(self, node)
    self.body.append(CR + r'\sphinxstepscope' + CR)
  def visit_footnote_patched(self, node: Element) -> None:
    visit_footnote_orig(self, node)
    if 'auto' in node:
      label = cast(nodes.label, node[0])
      self.body.append(r'\phantomsection\label{\thesphinxscope.%s}%%' % label.astext() + CR)
  sphinx.writers.latex.LaTeXTranslator.visit_start_of_file = visit_start_of_file_patched
  sphinx.writers.latex.LaTeXTranslator.visit_footnote = visit_footnote_patched

def setup(app):
  pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2017, Ignacio Fdez. Galván                             *
#***********************************************************************

# This is a convenience script to generate include files for link_blas.f
# The files are generated from a list of procedures from LAPACK/BLAS

from __future__ import (unicode_literals, division, absolute_import, print_function)
import sys
sys.dont_write_bytecode = True

import re
import os.path

# Read the lists of function from blas.fh, lapack.fh and legacy_mod.f

blas_functions = []
with open('../Include/blas.fh', 'r') as f:
  for line in f:
    match = re.match('#include "(.*)\.[fF](90)?"', line)
    if (match):
      blas_functions.append(match.group(1))
    match = re.match('!.* defines (.*)', line)
    if (match):
      blas_functions.extend(match.group(1).split(', '))
blas_functions.sort()

lapack_functions = []
with open('../Include/lapack.fh', 'r') as f:
  for line in f:
    match = re.match('#include "(.*)\.[fF](90)?"', line)
    if (match):
      lapack_functions.append(match.group(1))
    match = re.match('!.* defines (.*)', line)
    if (match):
      lapack_functions.extend(match.group(1).split(', '))
lapack_functions.sort()

legacy_functions = []
with open('legacy_mod.f', 'r') as f:
  for line in f:
    match = re.match('\s*subroutine (.*)\(', line)
    if (match):
      legacy_functions.append(match.group(1))
    match = re.match('!.* defines (.*)', line)
legacy_functions.sort()

# Generate the include files f1.fh, f2.fh, f3.fh, f4.fh, f5.fh

header = '''\
!***********************************************************************
! This file is part of OpenMolcas.                                     *
!                                                                      *
! OpenMolcas is free software; you can redistribute it and/or modify   *
! it under the terms of the GNU Lesser General Public License, v. 2.1. *
! OpenMolcas is distributed in the hope that it will be useful, but it *
! is provided "as is" and without any express or implied warranties.   *
! For more details see the full text of the license in the file        *
! LICENSE or in <http://www.gnu.org/licenses/>.                        *
!***********************************************************************
! File generated with {0}\n
'''.format(os.path.basename(__file__))

with open('f1.fh', 'w') as f:
  f.write(header)
  f.write('  use blas_mod')
  for i in blas_functions:
    f.write(', &\n      int_{0}=>{0}'.format(i))
  f.write('\n')
  f.write('  use lapack_mod')
  for i in lapack_functions:
    f.write(', &\n      int_{0}=>{0}'.format(i))
  f.write('\n')
  f.write('  use legacy_mod')
  for i in legacy_functions:
    f.write(', &\n      int_{0}=>{0}'.format(i))
  f.write('\n')

with open('f2.fh', 'w') as f:
  f.write(header)
  f.write('! BLAS procedures\n')
  for i in blas_functions:
    f.write('  procedure(int_{0}), pointer :: lb_{0}\n'.format(i))
  f.write('! LAPACK procedures\n')
  for i in lapack_functions:
    f.write('  procedure(int_{0}), pointer :: lb_{0}\n'.format(i))
  f.write('! Legacy procedures\n')
  for i in legacy_functions:
    f.write('  procedure(int_{0}), pointer :: lb_{0}\n'.format(i))

with open('f3.fh', 'w') as f:
  f.write(header)
  f.write('''!
!     BLAS procedures\n''')
  for i in blas_functions:
    f.write('''!
      funptr=link_func('{0}')
      if (c_associated(funptr)) then
        call c_f_procpointer(funptr, lb_{0})
      end if\n'''.format(i))
  f.write('''!
!     LAPACK procedures\n''')
  for i in lapack_functions:
    f.write('''!
      funptr=link_func('{0}')
      if (c_associated(funptr)) then
        call c_f_procpointer(funptr, lb_{0})
      end if\n'''.format(i))
  f.write('''!
!     Legacy procedures\n''')
  for i in legacy_functions:
    f.write('''!
      funptr=link_func('{0}')
      if (c_associated(funptr)) then
        call c_f_procpointer(funptr, lb_{0})
      end if\n'''.format(i))

with open('f4.fh', 'w') as f:
  f.write(header)
  f.write('''      !
      ! BLAS\n''')
  for i in blas_functions:
    f.write('      lb_{0}=>int_{0}\n'.format(i))
  f.write('''      !
      ! LAPACK\n''')
  for i in lapack_functions:
    f.write('      lb_{0}=>int_{0}\n'.format(i))
  f.write('''      !
      ! Legacy\n''')
  for i in legacy_functions:
    f.write('      lb_{0}=>int_{0}\n'.format(i))

with open('f5.fh', 'w') as f:
  f.write(header)
  f.write('''      ! BLAS
      !\n''')
  for i in blas_functions:
    f.write('''      if (DLAddr(c_funloc(lb_{0}),c_loc(info)) /= 0) then
        write(6,*) '{0} from: ',c_f_string(info%dli_fname)
      else
        write(6,*) 'no {0} found!'
      end if\n'''.format(i))
  f.write('''
      ! LAPACK
      !\n''')
  for i in lapack_functions:
    f.write('''      if (DLAddr(c_funloc(lb_{0}),c_loc(info)) /= 0) then
        write(6,*) '{0} from: ',c_f_string(info%dli_fname)
      else
        write(6,*) 'no {0} found!'
      end if\n'''.format(i))
  f.write('''
      ! Legacy
      !\n''')
  for i in legacy_functions:
    f.write('''      if (DLAddr(c_funloc(lb_{0}),c_loc(info)) /= 0) then
        write(6,*) '{0} from: ',c_f_string(info%dli_fname)
      else
        write(6,*) 'no {0} found!'
      end if\n'''.format(i))



# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015-2018, Ignacio Fdez. Galván                        *
#***********************************************************************

from __future__ import (unicode_literals, division, absolute_import, print_function)
try:
  from six import text_type
except ImportError:
  text_type = str

from os import environ, getcwd, symlink, access, F_OK, X_OK, pathsep, defpath, curdir
from os.path import join, split, isfile, isdir, exists, expanduser, realpath, dirname, normcase
from sys import platform, executable
from tempfile import mkdtemp
from shutil import rmtree
from re import sub, match
from subprocess import Popen, PIPE
from json import loads

def set_utf8(var, val, dummy_val='UNKNOWN_VARIABLE'):
  '''Auxiliary function to set environment variables
     making sure they are utf-8-encoded (works with python 2 and 3).
  '''
  try:
    environ[var] = val
  except TypeError:
    environ[var] = text_type(val)
  except UnicodeEncodeError:
    environ[var] = val.encode('utf-8')
  if (val == dummy_val):
    del environ[var]

def get_utf8(var, default=None, dummy_val='UNKNOWN_VARIABLE'):
  '''Auxiliary function to get environment variables
     making sure they are utf-8-encoded (works with python 2 and 3).
  '''
  try:
    val = environ.get(var, default).decode('utf-8')
  except AttributeError:
    val = environ.get(var, default)
  if (val == dummy_val):
    return default
  else:
    return val

def expandvars(string, default=None, skip_escaped=False):
  '''Expand environment variables of form $var and ${var}.
     If parameter 'skip_escaped' is True, all escaped variable references
     (i.e. preceded by backslashes) are skipped.
     Unknown variables are set to 'default'.
     If 'default' is None, they are not expanded.
  '''
  def replace_var(m):
    return environ.get(m.group(2) or m.group(1), m.group(0) if default is None else default)
  reVar = (r'(?<!\\)' if skip_escaped else '') + r'\$(\w+|\{([^}]*)\})'
  return sub(reVar, replace_var, string)

def utf8_open(*args, **kwargs):
  '''Open files using io.open, with utf-8 encoding by default (not in binary mode)'''
  import io
  if ((len(args) > 1) and ('b' in args[1])):
    return io.open(*args, **kwargs)
  else:
    return io.open(*args, encoding='utf-8', **kwargs)

def dotmolcas(filename):
  return join(expanduser('~'), '.Molcas', filename)

def find_molcas(xbin_list=None, here=True):
  '''Find a molcas installation and define MOLCAS
     based on a tag file or the path in some configuration file.
     Also add the molcas libraries to LD_LIBRARY_PATH
  '''
  # walk up the tree to find .molcashome
  # (this overwrites MOLCAS if defined)
  if (here):
    path = getcwd()
    (head, tail) = split(path)
    if (tail == ''):
      (head, tail) = split(head)
    while (tail != ''):
      if (isfile(join(path, '.molcashome'))):
        set_utf8('MOLCAS', path)
        break
      path = head
      (head, tail) = split(path)

  # look up pre-defined locations
  pathfiles = [dotmolcas('molcas_default'), dotmolcas('molcas'), join('/etc', 'alternatives', 'molcas')]
  if (get_utf8('MOLCAS', default='') == ''):
    for fn in pathfiles:
      if (isfile(fn)):
        with utf8_open(fn, 'r') as f:
          set_utf8('MOLCAS', f.read().strip())
        break

  # if MOLCAS is still not defined, try with the driver's location
  if (get_utf8('MOLCAS', default='') == ''):
    from __main__ import __file__ as main_file
    path = dirname(realpath(main_file))
    if (isfile(join(path, '.molcashome'))):
      set_utf8('MOLCAS', path)

  path = get_utf8('MOLCAS', default='')

  # set up LD_LIBRARY_PATH
  if (path != ''):
    llp = get_utf8('LD_LIBRARY_PATH', default='')
    lp = join(path, 'lib')
    if (llp == ''):
      set_utf8('LD_LIBRARY_PATH', lp)
    else:
      set_utf8('LD_LIBRARY_PATH', '{0}:{1}'.format(lp, llp))

  # process custom environment commands
  filename = dotmolcas('molcas.shell')
  if isfile(filename):
    source = '. {0}'.format(filename)
    dump = '{0} -c "import os, json;print(json.dumps(dict(os.environ)))"'.format(executable)
    pipe = Popen(['/bin/sh', '-c', '{0} && {1}'.format(source, dump)], stdout=PIPE)
    try:
      env = loads(pipe.stdout.read().decode('utf-8'))
      environ.update(env)
    except:
      print('Warning: {0} could not be processed\n'.format(filename))

  # read xbin.cfg for custom executables
  # (no expansion here)
  if (xbin_list is not None):
    filename = dotmolcas('xbin.cfg')
    if isfile(filename):
      with utf8_open(filename, 'r') as xf:
        for line in xf:
          a, b = line.split('=')
          xbin_list[a.strip()] = b.strip()

def find_sources():
  '''Find the source directories from a build directory,
     if using CMake, and sets the *_SOURCE variables.
     Otherwise they default to $MOLCAS
  '''
  MOLCAS = get_utf8('MOLCAS')
  cmake_src = ''
  OPENMOLCAS_SOURCE = ''
  MOLCAS_SOURCE = ''
  try:
    with utf8_open(join(MOLCAS, 'CMakeCache.txt'), 'r') as f:
      for line in f:
        m = match(r'OpenMolcas_SOURCE_DIR:.*?=(.*)', line)
        if (m):
          cmake_src = m.group(1)
        m = match(r'OPENMOLCAS_DIR:.*?=(.*)', line)
        if (m):
          OPENMOLCAS_SOURCE = m.group(1)
        m = match(r'EXTRA:.*?=(.*)', line)
        if (m):
          MOLCAS_SOURCE = m.group(1)
  except:
    pass
  if (cmake_src == ''):
    cmake_src = MOLCAS
  if (MOLCAS_SOURCE == ''):
    MOLCAS_SOURCE = cmake_src
  if (OPENMOLCAS_SOURCE == ''):
    OPENMOLCAS_SOURCE = cmake_src
  if (get_utf8('MOLCAS_SOURCE', default='') == ''):
    set_utf8('MOLCAS_SOURCE', MOLCAS_SOURCE)
  if (get_utf8('OPENMOLCAS_SOURCE', default='') == ''):
    set_utf8('OPENMOLCAS_SOURCE', OPENMOLCAS_SOURCE)

def attach_streams(output, error, buffer_size=-1):
  '''Attach output and error streams to files, with optional buffer size'''
  import sys
  if (output is not None):
    sys.stdout = utf8_open(output, 'w', buffer_size)
  if (error is not None):
    if (error == output):
      sys.stderr = sys.stdout
    else:
      sys.stderr = utf8_open(error, 'w', buffer_size)

def which(cmd, mode=F_OK | X_OK, path=None):
  '''Copy of shutil.which, since it is not available for python < 3.3
     Modified to return realpath'''
  def _access_check(fn, mode):
    return (exists(fn) and access(fn, mode) and not isdir(fn))
  if dirname(cmd):
    if _access_check(cmd, mode):
      return realpath(cmd)
    return None
  if path is None:
    path = environ.get('PATH', defpath)
  if not path:
    return None
  path = path.split(pathsep)
  if platform == 'win32':
    if curdir not in path:
      path.insert(0, curdir)
    pathext = environ.get('PATHEXT', '').split(pathsep)
    if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
      files = [cmd]
    else:
      files = [cmd + ext for ext in pathext]
  else:
    files = [cmd]
  seen = set()
  for dir in path:
    normdir = normcase(dir)
    if normdir not in seen:
      seen.add(normdir)
      for thefile in files:
        name = join(dir, thefile)
        if _access_check(name, mode):
          return realpath(name)
  return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2019, Ignacio Fdez. Galván                             *
#***********************************************************************
#
# Split a list of tests in sets taking roughly the same running time
#
# ./split_tests.py flatlist total_number selected_number

from __future__ import (division, print_function)
import sys, os, re

# default running time for tests with no estimate
default_time = 60

# this file location
path = os.path.dirname(os.path.abspath(__file__))

# flatlist
try:
  flatlist = sys.argv[1]
except IndexError:
  sys.exit()

# total sets
try:
  numsets = int(sys.argv[2])
  assert numsets > 0
except:
  numsets = 1

# this set
try:
  thisnum = int(sys.argv[3])
  assert 0 < thisnum <= numsets
except:
  thisnum = 1

# read running time estimates
timest = {}
try:
  with open(os.path.join(path, 'timing.data'), 'r') as f:
    while True:
      try:
        line = next(f)
        match = re.match(r'--- (\S*)( .)?', line)
        if match:
          test = match.group(1)
          timest[test] = float(next(f))
          if match.group(2):
            timest[test] = max(timest[test], default_time)
      except StopIteration:
        break
except IOError:
  pass

# read test list
testlist = []
with open(flatlist, 'r') as f:
  for l in f:
    testlist.append(l.strip())

# compute total time
total = 0
for i in testlist:
  total += timest.get(i, default_time)

# split in sets
set_time = total / numsets
sets = []
this_set = []
this_time = 0
for i in testlist:
  this_set.append(i)
  this_time += timest.get(i, default_time)
  if this_time > set_time:
    sets.append(this_set[:])
    this_set = []
    this_time = 0
if (len(sets) < numsets):
  sets.append(this_set[:])

# print selected set
nothing = True
try:
  for i in sets[thisnum-1]:
    nothing = False
    print(i)
except IndexError:
  pass
if nothing:
  print ('.none')



#!/usr/bin/env python3

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2021, Ignacio Fdez. Galván                             *
#***********************************************************************

# Help script for reformatting a basis set file in Molcas format.
# It writes the numbers in a consistent format and fails if precision
# is lost.
# Use with compact=True and remove_comments=True to facilitate comparison
# between different versions.

import sys
import re
import numpy as np
from decimal import Decimal

# Output formats: (total_places, decimal_places)
#           if compact, just use minimal space
compact  =  False
#           format for basis set exponents
exp_form =  [20, 8]
#           format for basis set coefficients, orbital energies and Fock operator
coe_form =  [15, 12]
#           format for pseudopotential parameters
pp_form =   [14, 8]
#           format for "M1" and "M2" parameters
ecpm_form = [15, 12]
#           format for "PROJOP" exponents
ecpe_form = [20, 8]
#           format for "PROJOP" coefficients and constants
ecpc_form = [15, 12]
#           maximum line length
max_line =  180

uncontracted = False
remove_comments = False
insert_comments = False

l_lab = 'spdfghi'

if compact:
  exp_form[0] = None
  coe_form[0] = None
  pp_form[0] = None
  ecpm_form[0] = None
  ecpe_form[0] = None
  ecpc_form[0] = None

# Read next line, ignores (and prints unchanged) comments and empty lines
def nextline(f):
  global uncontracted
  line = f.readline()
  while line and (line.startswith('*') or line.startswith('#') or line.startswith('!') or line.startswith('\n')):
    if re.match(r'#Contraction *UNC', line, re.I):
      uncontracted = True
    if not remove_comments:
      print(line.rstrip())
    line = f.readline()
  if line:
    return line.rstrip()
  else:
    return None

def count_items(parts):
  n = 0
  for i in parts:
    if i.startswith('*'):
      break
    n += 1
  return n

def read_nums(f, n):
  nums = []
  while len(nums) < n:
    line = nextline(f)
    nums.extend(line.split())
    if nums[-1] == '/':
      l = n-len(nums)+1
      nums = nums[:-1]+l*['0']
  assert(len(nums) == n)
  return nums

def print_line(nums, form):
  parts = [f' {format_num(i, *form)}' for i in nums]
  if max_line is None:
    print(''.join(parts))
  else:
    line = ''
    for i in parts:
      if len(line) + len(i) > max_line:
        print(line)
        line = ''
      line += i
    print(line)

################################################################################
# Set of functions to "pretty print" a number in some reserved space,
# ensuring that no precision is lost.
# Quite hackish and inefficient, but as long as it works...

# Return the "scale" of a number, i.e, the exponent of its leading digit
def scale(num):
  if num == 0:
    return 0
  return num.adjusted()

# Significant places of a number
def sig_places(num):
  digits = list(num.as_tuple().digits)
  while len(digits) > 1:
    if digits[-1] != 0:
      break
    digits.pop()
  return len(digits)

# Try to fit a number in n total places with m decimals.
# Returns a string of asterisks if it fails
def fit_num(f, n, m):
  s = sig_places(f)
  e = scale(f)
  sg = 1 if f < 0 else 0
  if (s-e-1 <= m) and (e+1 <= n-m-sg-1):
    out = f'{f:{n}.{m}f}'
  else:
    out = '*'*n
  return out

# Try to fit a number in n total places with m decimals, with a exponent (shift)
# Returns a string of asterisks if it fails
def print_shifted(f, n, m, shift=0):
  fs = f.scaleb(shift)
  if shift == 0:
    es = 0
  else:
    es = np.int(np.floor(np.log10(abs(shift))))+2
    if shift > 0:
      es += 1
  s = sig_places(fs)
  if shift == 0:
    out = fit_num(fs, n, m)
  else:
    out = fit_num(fs, n-es, m-es)
    out += f'e{-shift}'
  if len(out) > n or out[0] == '*':
    out = '*'*n
  return out

# Simplify a number that may come from an "overprecise" float,
# i.e. round to fewer decimals as long as it represents the same float number
def simplify(f):
  out = f
  ref = float(f)
  s, n, e = f.as_tuple()
  m = len(n)
  for i in range(m):
    a = f.quantize(f.scaleb(i+1))
    if float(a) != ref:
      break
    out = a
  if out != f:
    print(f'simplify {f} to {out}', file=sys.stderr)
  return out

# Regexp to fix exponents from Fortran prints, which may have D instead of E
fortfixexp = re.compile(r'([\d.])[dD]?(((?<=[dD])[+-]?|[+-])\d)')

# Format a number using the minumum necessary space
# (but still using at least one explicit zero before/after decimal point)
def format_minimal(num):
  t = num.as_tuple()
  d = [i for i in t.digits]
  e = t.exponent
  # remove trailing zeros
  while (len(d) > 1) and d[-1] == 0:
    d.pop()
    e += 1
  # "text" contains all significant digits
  text = ''.join([f'{i}' for i in d])
  # special case for zero
  if text == '0':
    return text
  # large number, use exponent if large enough
  if ((len(d) > 1) and (e > 3)) or (e > 4):
    if len(d) == 1:
      text += f'.0e{e}'
    else:
      s = len(d)-1
      # prefer fewer digits in exponent
      if (e+s > 9) and (e+1 < 10):
        s = 9-e 
      elif (e+s > 99) and (e+1 < 100):
        s = 99-e
      e_ = e+s
      text = text[0:-s] + '.' + text[-s:] + f'e{e_}'
  # not too large, not too small, print without exponent
  elif e >= 0:
    text += e*'0'
  elif len(d)+e > 0:
    text = text[:e] + '.' + text[e:]
  elif ((len(d) > 1) and (-e-len(d) < 3)) or (e > -5):
    text = '0.' + (-len(d)-e)*'0' + text
  # small number, use exponent
  else:
    if len(d) == 1:
      s = -2
      e_ = e
      z = '0'
    else:
      s = len(d)-1
      e_ = e+s
      z = ''
    text = text[0:-s] + '.' + text[-s:] + z + f'e{e_}'
  out = '-' if t.sign else ''
  out += text
  return out

# Format a decimal number in n total places with m decimals, adjusting
# the exponent if necessary.
# If n = None, formats the number with the minimal space needed.
# Raises exception if failed.
def format_num(num, n=None, m=None):
  num = fortfixexp.sub(r'\1e\2', num)
  f = Decimal(num)
  f = simplify(f)
  if n is None:
    out = format_minimal(f)
  else:
    out = print_shifted(f, n, m)
    # if it doesn't fit without exponent, try shifting until it fits
    if out[0] == '*':
      e = scale(f)
      if e > 0:
        for i in range(-e, 0):
          out = print_shifted(f, n, m, shift=i)
          if out[0] != '*':
            break
      else:
        for i in range(-e, -e+(n-m-1)):
          out = print_shifted(f, n, m, shift=i)
          if out[0] != '*':
            break
  if out[0] == '*':
    raise Exception(f'Cannot fit number in ({n}.{m}): {num}')
  # ensure that the final number represents the same float
  if float(num) != float(out):
    raise Exception(f'Changed value in ({n}.{m}): {num} -> {out}')
  return out
################################################################################

with open(sys.argv[1]) as f:
  line = nextline(f)
  while line is not None:
    if line.startswith('/'):
      # basis label
      print(line)
      parts = line.split('.')
      prim = parts[3]
      prim = re.sub(r'[spdfghi]', ' ', prim).split()
      n_p = len(prim)-1
      n_prim = [int(i) for i in prim]
      func = parts[4]
      func = re.sub(r'[spdfghi]', ' ', func).split()
      n_f = [int(i) for i in func]
      if uncontracted:
        assert(np.all(prim == func))
      # two reference lines
      line = nextline(f)
      print(line)
      line = nextline(f)
      print(line)
      options = ''
      line = nextline(f)
    elif re.match(r'Options', line, re.I):
      # all lines until "EndOptions"
      print(line)
      line = nextline(f)
      while line is not None and not re.match(r'End([ O]|$)', line, re.I):
        print(line)
        if re.match(r'FockOperator', line, re.I):
          options += 'f'
        if re.match(r'OrbitalEnergies', line, re.I):
          options += 'e'
        line = nextline(f)
      print(line)
      line = nextline(f)
    elif re.match(r'Spectral', line, re.I):
      # all lines until "End of Spectral"
      print(line)
      line = nextline(f)
      while line is not None and not re.match(r'End([ o]|$)', line, re.I):
        print(line)
        if re.match(r'Exte', line, re.I):
          line = nextline(f)
          maxl = int(line)
          print(f' {maxl:4d}')
          for i in range(maxl+1):
            line = nextline(f)
            mp, = [int(j) for j in line.split()]
            print(f' {mp:4d}')
            exps = read_nums(f, mp)
            for e in exps:
              print(format_num(e, *exp_form))
        elif line == 'SOC':
          line = nextline(f)
          maxl = int(line)
          print(f' {maxl:4d}')
          for i in range(maxl+1):
            line = nextline(f)
            mp, mo, md = [int(j) for j in line.split()]
            if insert_comments:
              print(f'* {l_lab[i]}-type functions')
            print(f' {mp:4d} {mo:4d} {md:4d}')
            exps = read_nums(f, mp)
            for e in exps:
              print(format_num(e, *exp_form))
            for c in range(mp):
              coefs = read_nums(f, mo)
              print_line(coefs, coe_form)
        line = nextline(f)
      print(line)
      line = nextline(f)
    elif re.match(r'PP', line, re.I):
      parts = [i.strip() for i in line.strip(';').split(',')]
      parts[0] = 'PP'
      parts[1] = parts[1][0].upper() + parts[1][1:].lower()
      maxl = int(parts[3])
      print(', '.join(parts)+' ;')
      for nl in range(maxl+1):
        line = nextline(f).split(';')[0]
        n = int(line)
        print(f'{n:3d} ;')
        for m in range(n):
          line = nextline(f).strip(';')
          parts = line.split(',')
          i = int(parts[0])
          f1 = parts[1]
          f2 = parts[2]
          print(f'{i:2d}, {format_num(f1, *pp_form)}, {format_num(f2, *pp_form)} ;')
      line = nextline(f)
    elif re.match(r'M[12]', line, re.I):
      print(line.strip().upper())
      line = nextline(f)
      n = int(line)
      print(f'{n:3d}')
      if n > 0:
        for i in [1, 2]:
          nums = read_nums(f, n)
          print_line(nums, ecpm_form)
      line = nextline(f)
    elif re.match(r'COREREP', line, re.I):
      print('COREREP')
      line = nextline(f)
      print(line)
      line = nextline(f)
    elif re.match(r'PROJOP', line, re.I):
      print('PROJOP')
      line = nextline(f)
      maxl = int(line)
      print(f' {maxl:4d}')
      for i in range(maxl+1):
        line = nextline(f)
        parts = [int(j) for j in line.split()]
        mp = parts[0]
        mo = parts[1]
        occ = parts[2:]
        assert(len(occ) <= mo)
        if len(occ) == mo:
          if all([j == 4*i+2 for j in occ]):
            occ = []
        print(f' {mp:4d} {mo:4d}' + ''.join([f' {j:4d}' for j in occ]))
        coefs = read_nums(f, mo)
        print_line(coefs, ecpc_form)
        exps = read_nums(f, mp)
        for e in exps:
          print(format_num(e, *ecpe_form))
        for c in range(mp):
          coefs = read_nums(f, mo)
          print_line(coefs, ecpc_form)
      line = nextline(f)
    else:
      # here go the exponents and coefficients for the basis set
      parts = line.split()
      Z = float(parts[0])
      maxl = n_p
      print(f'   {Z:5.1f} {maxl:3d}')
      line = nextline(f)
      for nl in range(maxl+1):
        if insert_comments:
          print(f'* {l_lab[nl]}-type functions')
        parts = line.split()
        n_p = int(parts[0])
        assert(n_prim[nl] == n_p)
        assert(count_items(parts) <= 2)
        if count_items(parts) > 1:
          assert(n_f[nl] == int(parts[1]))
        print(f' {n_p:4d} {n_f[nl]:4d}')
        exps = read_nums(f, n_p)
        for e in exps:
          print(format_num(e, *exp_form))
        if not uncontracted:
          for c in range(n_p):
            coefs = read_nums(f, n_f[nl])
            print_line(coefs, coe_form)
        line = nextline(f)
        if 'f' in options:
          n = int(line)
          print(f' {n:4d}')
          for c in range(n):
            coefs = read_nums(f, n)
            print_line(coefs, coe_form)
          line = nextline(f)
        if 'e' in options:
          n = int(line)
          print(f' {n:4d}')
          coefs = read_nums(f, n)
          if n:
            print_line(coefs, coe_form)
          line = nextline(f)



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Oskar Weser                                      *
#***********************************************************************

#  """
#  Entrypoint module, in case you use `python -manalyze_molcas`.
#
#
#  Why does this file exist, and why __main__? For more info, read:
#
#  - https://www.python.org/dev/peps/pep-0338/
#  - https://docs.python.org/2/using/cmdline.html#cmdoption-m
#  - https://docs.python.org/3/using/cmdline.html#cmdoption-m
#  """
#  from analyze_molcas.cli import main
#
#  if __name__ == "__main__":
#      main()



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Oskar Weser                                      *
#***********************************************************************

from typing import Sequence, Tuple, TypeVar, Type, TextIO
import re
from os import PathLike
from enum import Enum
from copy import deepcopy
from abc import ABCMeta, abstractmethod

import numpy as np
from numpy import array


class FileFormat(Exception):
    pass


class RasOrb_version(Enum):
    v2_0, v2_2 = range(2)


T = TypeVar('T', bound='_Orbitals')


class _Orbitals(metaclass=ABCMeta):

    def __init__(self, orbs, coeff, occ, energy, idx):
        self.orbs = orbs
        self.coeff = coeff
        self.occ = occ
        self.energy = energy
        self.idx = idx

    @classmethod
    @abstractmethod
    def read_orbfile(cls: Type[T], path: PathLike) -> T:
        pass

    @abstractmethod
    def reindex(self,
                new_idx: Sequence[Sequence[int]], inplace: bool=False) -> T:
        pass

    def get_index(self) -> Sequence[Sequence[int]]:
        return [np.arange(n_orbs) for n_orbs in self.orbs]

    def copy(self) -> T:
        return self.__class__(
            orbs=deepcopy(self.orbs),
            coeff=deepcopy(self.coeff),
            occ=deepcopy(self.occ),
            energy=deepcopy(self.energy),
            idx=deepcopy(self.idx))

    def _write_orbfile_stream(self):
        yield from self._write_header()
        yield from self._write_MO_coeff()
        yield from self._write_MO_occ()
        yield from self._write_MO_human_occ()
        yield from self._write_MO_E()
        yield from self._write_CAS_idx()

    def write_orbfile(self, path=None):
        "Write an orbital file. By default an iterator of lines is returned."
        if path is None:
            return self._write_orbfile_stream()
        else:
            with open(path, 'w') as f:
                for line in self._write_orbfile_stream():
                    print(line, file=f)

    @abstractmethod
    def _write_header(self):
        pass

    @abstractmethod
    def _write_MO_coeff(self):
        pass

    @abstractmethod
    def _write_MO_occ(self):
        pass

    @abstractmethod
    def _write_MO_human_occ(self):
        pass

    @abstractmethod
    def _write_MO_E(self):
        pass

    def _write_CAS_idx(self):
        def res_out(v):
            lines = _reshape_output(v, 10, '1')
            return (f'{i % 10} {a}'for i, a in enumerate(lines))

        lines = ['#INDEX']
        for irrep, n_orbs in enumerate(self.orbs):
            lines.append('* 1234567890')
            lines.extend(res_out(self.idx[irrep]))
        return lines

    def __repr__(self):
        return ("{} with {} orbitals per irrep"
                .format(type(self).__name__, self.orbs))


def _reindex(indices: Sequence[Sequence],
             new_indices: Sequence[Sequence[int]]) -> Sequence[Sequence]:
    return [idx[new_idx] for new_idx, idx in zip(new_indices, indices)]


def _spin_reindex(values: Sequence[Sequence],
                  new_indices: Sequence[Sequence[int]]) -> Sequence[Sequence]:
    return {
        spin:
            [v[new_idx] for new_idx, v in zip(new_indices, spin_values)]
        for spin, spin_values in values.items()}


def _read_header(f: TextIO) -> Tuple[RasOrb_version, Sequence[int], bool]:
    line = f.readline()
    while line:
        if '#INPORB' in line:
            version = _read_version(line)
        elif '#INFO' in line:
            spin_orbs, orbs = _read_info(f)
            break
        line = f.readline()
    return version, orbs, spin_orbs


def _read_spin_orbs(
        f: TextIO, orbs: Sequence[int],
        version: RasOrb_version) -> Tuple[Sequence[array], ...]:
    MO_coeff, MO_occ, MO_E = {}, {}, {}
    line = f.readline()
    while line:
        if '#ORB' in line:
            MO_coeff['a'] = _read_coeffs(f, orbs, 5)
        elif '#UORB' in line:
            MO_coeff['b'] = _read_coeffs(f, orbs, 5)
        elif '#OCC' in line:
            MO_occ['a'] = _read_orb_property(
                f, orbs, 10 if version == RasOrb_version.v2_0 else 5)
        elif '#UOCC' in line:
            MO_occ['b'] = _read_orb_property(
                f, orbs, 10 if version == RasOrb_version.v2_0 else 5)
        elif '#ONE' in line:
            MO_E['a'] = _read_orb_property(f, orbs, 10)
        elif '#UONE' in line:
            MO_E['b'] = _read_orb_property(f, orbs, 10)
        elif '#INDEX' in line:
            idx = _read_CAS_idx(f, orbs, 10)
        line = f.readline()
    return MO_coeff, MO_occ, MO_E, idx


def _read_spatial_orbs(
        f: TextIO, orbs: Sequence[int],
        version: RasOrb_version) -> Tuple[Sequence[array], ...]:
    line = f.readline()
    while line:
        if '#ORB' in line:
            MO_coeff = _read_coeffs(f, orbs, 5)
        elif '#OCC' in line:
            MO_occ = _read_orb_property(
                f, orbs, 10 if version == RasOrb_version.v2_0 else 5)
        elif '#ONE' in line:
            MO_E = _read_orb_property(f, orbs, 10)
        elif '#INDEX' in line:
            idx = _read_CAS_idx(f, orbs, 10)
        line = f.readline()
    return MO_coeff, MO_occ, MO_E, idx


def _read_version(line: str) -> RasOrb_version:
    match = re.match('#INPORB (2).([02])', line)
    if match:
        major, minor = match.group(1, 2)
        if major == '2':
            if minor == '0':
                return RasOrb_version.v2_0
            if minor == '2':
                return RasOrb_version.v2_2
        raise ValueError(f'Unknown version {major}.{minor}')
    else:
        raise ValueError('Inporb version could not be determined.')


def _read_info(f: TextIO) -> Tuple[bool, Sequence[int]]:
    _forward(f, 1)
    line = f.readline()
    spin_orbs, n_sym, _ = (int(x) for x in line.split())
    line = f.readline()
    return bool(spin_orbs), [int(irrep) for irrep in line.split()[:n_sym]]


def _read_coeffs(
        f: TextIO, orbs: Sequence[int], cols: int) -> Sequence[array]:
    MO_coeff = [np.empty((n_orbs, n_orbs)) for n_orbs in orbs]
    for irrep, n_orbs in enumerate(orbs):
        rows = (n_orbs + cols - 1) // cols
        for orb in range(n_orbs):
            values = []
            _forward(f, 1)
            for _ in range(rows):
                line = f.readline()
                values.extend([float(x) for x in line.split()])
            MO_coeff[irrep][:, orb] = values
    return MO_coeff


def _read_orb_property(
        f: TextIO, orbs: Sequence[int], cols: int) -> Sequence[array]:
    MO_v = [np.empty(n_orbs) for n_orbs in orbs]
    f.readline()
    for irrep, n_orbs in enumerate(orbs):
        rows = (n_orbs + cols - 1) // cols
        values = []
        for _ in range(rows):
            line = f.readline()
            values.extend([float(x) for x in line.split()])
        MO_v[irrep][:] = values
    return MO_v


def _read_CAS_idx(
        f: TextIO, orbs: Sequence[int], cols: int) -> Sequence[array]:
    idx = []
    for n_orbs in orbs:
        rows = (n_orbs + cols - 1) // cols
        values = []
        next(f)
        for _ in range(rows):
            line = f.readline()
            values.extend(line.strip()[2:])
        idx.append(array(values))
    return idx


def _reshape_output(v, cols, formatter):
    rows = (len(v) + cols - 1) // cols
    for i in range(rows - 1):
        current = v[(i * cols):((i + 1) * cols)]
        yield ''.join(f'{x:{formatter}}' for x in current)
    current = v[(rows - 1) * cols:]
    yield ''.join(f'{x:{formatter}}' for x in current)


def _write_section_1D(title, orbs, values,
                      cols=5, subtitle=None, fmt='22.14E'):
    yield title
    if subtitle is not None:
        yield subtitle
    for irrep in range(len(orbs)):
        yield from _reshape_output(values[irrep][:], cols, fmt)


def _write_section_2D(title, orbs, values, cols=5, subtitle=None,
                      paragraph=None, fmt='22.14E'):
    yield title
    if subtitle is not None:
        yield subtitle
    for irrep, n_orbs in enumerate(orbs):
        for orb in range(n_orbs):
            if paragraph:
                yield paragraph(irrep, orb)
            yield from _reshape_output(values[irrep][:, orb], cols, fmt)


def _forward(f: TextIO, n: int) -> str:
    for _ in range(n):
        line = f.readline()
    return line



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#***********************************************************************

from __future__ import print_function
import os.path
import sys
import re
import xml.etree.ElementTree as ET

basedir = 'xml'

# Read argument
try:
  argument = sys.argv[1]
except IndexError:
  argument = None

database = ET.parse(os.path.join(basedir,'index.xml'))
root = database.getroot()

# First create a list of files, containing the directory where each find is located.
file_list={}
# Iterate directory items
for diritem in root.iterfind('compound[@kind="dir"]'):
  dirname = diritem.find('name').text
  # Open the XML file for each directory
  dirxml = diritem.get('refid') + '.xml'
  dirtree = ET.parse(os.path.join(basedir,dirxml))
  # For each file in this directory, assign the director name
  for fileitem in dirtree.getroot().iter('innerfile'):
    file_list[fileitem.get('refid')] = dirname

# Create a dictionary linking each element with its parent
parent_map = {c:p for p in root.iter() for c in p}

# Build the dictionary of functions, containg the file where each function is defined
name_list={}
# Iterate all "function" members belonging to "file" compounds
for member in root.iterfind('compound[@kind="file"]/member[@kind="function"]'):
  name = member.find('name').text
  # Find the directory and filename
  parent = parent_map[member]
  filename = parent.find('name').text
  # Ignore .h files
  if (not re.search(r"\.h$",filename)):
    dirname = file_list[parent.get('refid')]
    filename = os.path.join(dirname,filename)
    # Add an entry to the dictionary
    name_list.setdefault(name,[]).append(filename)

#===============================================================================

# No argument, print the full list
if (argument == None):
  for (name,files) in sorted(name_list.items()):
    for filename in sorted(files):
      print("{}:{}".format(name,filename))

# -Conf, find conflicts
elif (re.match('-Conf',argument)):
  check_list = {}
  expr = re.compile(r"(clones|^ga|^ma)")
  for (name,files) in sorted(name_list.items()):
    # Ignore files matching the above pattern
    for filename in sorted(files,reverse=True):
      if expr.search(filename):
        files.remove(filename)
    if (len(files) > 1):
      print("Conflict: {} found in {}".format(name,", ".join(files)))

# Search name, print the names matching the argument
else:
  expr = re.compile(argument)
  for name in sorted(filter(expr.search,name_list)):
    for filename in sorted(name_list[name]):
      print("{}:{}".format(name,filename))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2014, Steven Vancoillie                                *
#               2020, Ignacio Fdez. Galván                             *
#***********************************************************************
#
# modulenames.py:
#
# extract module names from a test input and generate a proper test name for
# use with CMake (no spaces, special chars)
#
# Steven Vancoillie, Lund, 20th of June 2014
# Ignacio Fdez. Galván, 2020: Translated from Perl to Python

import sys
import os
import re

def usage():
  name = os.path.basename(__file__)
  print('''

 Usage: {} <filename>

        Program to generate a test name by concatenating the
        module names used in a test input given by filename.
'''.format(name))
  sys.exit(1)

try:
  filename = sys.argv[1]
except:
  usage()

module = re.compile('&(\w+)')

module_names = []

try:
  with open(filename, 'r') as f:
    for line in f:
      matches = module.findall(line)
      for name in matches:
        lc_name = name.lower()
        if lc_name != 'end':
          module_names.append(lc_name)
except IOError:
  sys.exit('Error: failed to open file {}'.format(filename))

output = '_'.join(sorted(set(module_names)))

print(output)
